#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging

import numpy as np
from matplotlib import pyplot as plt

from mud_examples.utils import check_dir

plt.rcParams['figure.figsize'] = 10, 10
plt.rcParams['font.size'] = 16


__author__ = "Mathematical Michael"
__copyright__ = "Mathematical Michael"
__license__ = "mit"

_logger = logging.getLogger(__name__)
_mpl_logger = logging.getLogger('matplotlib')
_mpl_logger.setLevel(logging.WARNING)


def experiment_measurements(fun, num_measurements,
                            sd, num_trials, seed=21):
    """
    Fixed sensors, varying how much data is incorporated into the solution.
    """
    experiments = {}
    solutions = {}
    for ns in num_measurements:
        _logger.debug(f'Measurement experiment. Num measurements: {ns}')
        discretizations = []
        estimates = []
        for t in range(num_trials):
            np.random.seed(seed + t)
            _d = fun(sd=sd, num_obs=ns)
            estimate = _d.estimate()
            discretizations.append(_d)
            estimates.append(estimate)
        experiments[ns] = discretizations
        solutions[ns] = estimates

    return experiments, solutions


def experiment_equipment(fun, num_measure,
                         sd_vals, num_trials,
                         reference_value):
    """
    Fixed number of sensors, varying the quality of equipment.
    """
    sd_err = []
    sd_var = []
    for sd in sd_vals:
        _logger.debug(f'Equipment Experiment. Std Dev: {sd}')
        temp_err = []
        for t in range(num_trials):
            _d = fun(sd=sd, num_obs=num_measure)
            estimate = _d.estimate()
            temp_err.append(np.linalg.norm(estimate - reference_value))
        sd_err.append(np.mean(temp_err))
        sd_var.append(np.var(temp_err))

    return sd_err, sd_var


def plot_experiment_equipment(tolerances, res, prefix, fsize=32, linewidth=5,
                              title="Variance of MUD Error", save=True):
    print("Plotting experiments involving equipment differences...")
    plt.figure(figsize=(10, 10))
    for _res in res:
        _prefix, _in, _rm, _re = _res
        regression_err_mean, slope_err_mean, \
            regression_err_vars, slope_err_vars, \
            sd_means, sd_vars, num_sensors = _re
        plt.plot(tolerances, regression_err_mean,
                 label=f"{_prefix:10s} slope: {slope_err_mean:1.4f}",
                 lw=linewidth)
        plt.scatter(tolerances, sd_means, marker='x', lw=20)

    plt.yscale('log')
    plt.xscale('log')
    plt.Axes.set_aspect(plt.gca(), 1)
    plt.ylim(2E-3, 2E-2)
    # plt.ylabel("Absolute Error", fontsize=fsize)
    plt.xlabel('Tolerance', fontsize=fsize)
    plt.legend()
    plt.title(f"Mean of MUD Error for N={num_sensors}", fontsize=1.25 * fsize)
    if save:
        fdir = ''.join(prefix.split('/')[::-1])
        check_dir(fdir)
        _logger.info("Saving equipment experiments: mean convergence.")
        plt.savefig(f'{prefix}_convergence_mud_std_mean.png',
                    bbox_inches='tight')
    else:
        plt.show()

    plt.figure(figsize=(10, 10))
    for _res in res:
        _prefix, _in, _rm, _re = _res
        regression_err_mean, slope_err_mean, \
            regression_err_vars, slope_err_vars, \
            sd_means, sd_vars, num_sensors = _re
        plt.plot(tolerances, regression_err_vars,
                 label=f"{_prefix:10s} slope: {slope_err_vars:1.4f}",
                 lw=linewidth)
        plt.scatter(tolerances, sd_vars, marker='x', lw=20)
    plt.xscale('log')
    plt.yscale('log')
    plt.ylim(2E-5, 2E-4)
    plt.Axes.set_aspect(plt.gca(), 1)
    # plt.ylabel("Absolute Error", fontsize=fsize)
    plt.xlabel('Tolerance', fontsize=fsize)
    plt.legend()
    plt.title(title, fontsize=1.25 * fsize)
    if save:
        _logger.info("Saving equipment experiments: variance convergence.")
        plt.savefig(f'{prefix}_convergence_mud_std_var.png',
                    bbox_inches='tight')
    else:
        plt.show()


def plot_experiment_measurements(res, prefix,
                                 fsize=32, linewidth=5,
                                 xlabel='Number of Measurements',
                                 save=True, legend=False):
    print("Plotting experiments involving increasing # of measurements.")
    plt.figure(figsize=(10, 10))
    for _res in res:
        _prefix, _in, _rm, _re = _res
        solutions = _in[-1]
        measurements = list(solutions.keys())
        regression_mean, slope_mean, \
            regression_vars, slope_vars, \
            means, variances = _rm
        plt.plot(measurements[:len(regression_mean)], regression_mean,
                 label=f"{_prefix:4s} slope: {slope_mean:1.4f}",
                 lw=linewidth)
        plt.scatter(measurements[:len(means)], means, marker='x', lw=20)
    plt.xscale('log')
    plt.yscale('log')
    plt.Axes.set_aspect(plt.gca(), 1)
    plt.ylim(0.9 * min(means), 1.3 * max(means))
    plt.ylim(2E-3, 2E-1)
    plt.xlabel(xlabel, fontsize=fsize)
    if legend:
        plt.legend(fontsize=fsize * 0.8)
    # plt.ylabel('Absolute Error in MUD', fontsize=fsize)
    title = "$\\mathrm{\\mathbb{E}}(|\\lambda^* - \\lambda^\\dagger|)$"  # noqa E501
    plt.title(title, fontsize=1.25 * fsize)
    if save:
        fdir = '/'.join(prefix.split('/')[:-1])
        check_dir(fdir)
        _logger.info("Saving measurement experiments: mean convergence.")
        plt.savefig(f'{prefix}_convergence_obs_mean.png', bbox_inches='tight')
    else:
        plt.show()

    plt.figure(figsize=(10, 10))
    for _res in res:
        _prefix, _in, _rm, _re = _res
        regression_mean, slope_mean, \
            regression_vars, slope_vars, \
            means, variances = _rm
        plt.plot(measurements[:len(regression_vars)], regression_vars,
                 label=f"{_prefix:4s} slope: {slope_vars:1.4f}",
                 lw=linewidth)
        plt.scatter(measurements[:len(variances)], variances,
                    marker='x', lw=20)
    plt.xscale('log')
    plt.yscale('log')
    plt.Axes.set_aspect(plt.gca(), 1)
#     if not len(np.unique(variances)) == 1:
#         plt.ylim(0.9 * min(variances), 1.3 * max(variances))
    plt.ylim(5E-6, 5E-4)
    plt.xlabel(xlabel, fontsize=fsize)
    if legend:
        plt.legend(fontsize=fsize * 0.8)
    # plt.ylabel('Absolute Error in MUD', fontsize=fsize)
    plt.title("$\\mathrm{Var}(|\\lambda^* - \\lambda^\\dagger|)$",
              fontsize=1.25 * fsize)
    if save:
        _logger.info("Saving measurement experiments: variance convergence.")
        plt.savefig(f'{prefix}_convergence_obs_var.png', bbox_inches='tight')
    else:
        plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
# import os
import sys

# from mud_examples.runner import setup_logging
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import cm
# from mud import __version__ as __mud_version__
from mud.funs import map_sol, mud_sol
from mud.norm import full_functional, norm_data, norm_input, norm_predicted
from scipy.linalg import null_space

# from mud_examples import __version__
from mud_examples.utils import make_2d_unit_mesh
from mud_examples.parsers import parse_args
# maybe should segment out these examples at some point.
import mud_examples.linear.models as models
from mud_examples.utils import check_dir

plt.rcParams['figure.figsize'] = 10, 10
plt.rcParams['font.size'] = 24

__author__ = "Mathematical Michael"
__copyright__ = "Mathematical Michael"
__license__ = "mit"

_logger = logging.getLogger(__name__)  # TODO: make use of this instead of print


def setup_logging(loglevel):
    """Setup basic logging

    Args:
      loglevel (int): minimum loglevel for emitting messages
    """
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(level=loglevel, stream=sys.stdout,
                        format=logformat, datefmt="%Y-%m-%d %H:%M:%S")


def main_dim(args):
    """
    Main entrypoint for High-Dim Linear Dimension Example
    """
    args = parse_args(args)
    setup_logging(args.loglevel)
    np.random.seed(args.seed)
#     example       = args.example
#     num_trials   = args.num_trials
#     fsize        = args.fsize
#     linewidth    = args.linewidth
#     seed         = args.seed
    # dim_input     = args.input_dim
#     save         = args.save
#     alt          = args.alt
#     bayes        = args.bayes
#     prefix       = args.prefix
#     dist         = args.dist

    presentation = False
    save = True

    if not presentation:
        plt.rcParams['mathtext.fontset'] = 'stix'
        plt.rcParams['font.family'] = 'STIXGeneral'
    fdir = 'figures/lin'
    check_dir(fdir)

    fsize = 42

    def numnonzero(x, tol=1E-4):
        return len(x[abs(x) < tol])

    # # Impact of Dimension for Various Choices of $\\Sigma_\text{init}$
    # We sequentially incorporate $D=1, \dots , P$ dimensions into our QoI map and study the 2-norm between the true value that was used to generate the data and the analytical MUD/MAP points.

    # dim_output = dim_input
    dim_input, dim_output = 100, 100
    seed = 12
    np.random.seed(seed)

    # from sklearn.datasets import make_spd_matrix as make_spd
    # from sklearn.datasets import make_sparse_spd_matrix as make_cov
    # cov = np.eye(dim_input)
    initial_cov = np.diag(np.sort(np.random.rand(dim_input))[::-1] + 0.5)

    plt.figure(figsize=(10, 10))
    initial_mean = np.zeros(dim_input).reshape(-1, 1)
    # initial_mean = np.random.randn(dim_input).reshape(-1,1)
    randA = models.randA_gauss  # choose which variety of generating map
    A, b = models.randP(dim_input, randA=randA)
    prefix = 'lin-dim-cov'
    alpha_list = [10**(n) for n in np.linspace(-3, 4, 8)]

    # option to fix A and perturb lam_ref

    lam_ref = np.random.randn(dim_input).reshape(-1, 1)
    d = A@lam_ref + b

    # %%time
    sols = compare_linear_sols_dim(lam_ref, A, b, alpha_list, initial_mean, initial_cov)

    # c = np.linalg.cond(A)*np.linalg.norm(lam_ref)
    c = np.linalg.norm(lam_ref)
    # c = 1
    err_mud_list = [[np.linalg.norm(_m[0] - lam_ref) / c for _m in sols[alpha]] for alpha in alpha_list ]  # output_dim+1 values of _m
    err_map_list = [[np.linalg.norm(_m[1] - lam_ref) / c for _m in sols[alpha]] for alpha in alpha_list ]
    err_pin_list = [[np.linalg.norm(_m[2] - lam_ref) / c for _m in sols[alpha]] for alpha in alpha_list ]

    # c = np.linalg.cond(A)
    c = np.linalg.norm(A)
    err_Amud_list = [[np.linalg.norm(A @ (_m[0] - lam_ref)) / c for _m in sols[alpha]] for alpha in alpha_list ]
    err_Amap_list = [[np.linalg.norm(A @ (_m[1] - lam_ref)) / c for _m in sols[alpha]] for alpha in alpha_list ]
    err_Apin_list = [[np.linalg.norm(A @ (_m[2] - lam_ref)) / c for _m in sols[alpha]] for alpha in alpha_list ]

    # measure # of components that agree
    # err_mud_list = [[numnonzero(_m[0] - lam_ref) for _m in sols[alpha]] for alpha in alpha_list ]
    # err_map_list = [[numnonzero(_m[1] - lam_ref) for _m in sols[alpha]] for alpha in alpha_list ]
    # err_pin_list = [[numnonzero(_m[2] - lam_ref) for _m in sols[alpha]] for alpha in alpha_list ]

    x, y = np.arange(1, dim_output, 1), err_mud_list[0][0:-1]

    slope, intercept = (np.linalg.pinv(np.vander(x, 2))@np.array(y).reshape(-1, 1)).ravel()
    regression = slope * x + intercept


    # ---

    # # Convergence Plot

    for idx, alpha in enumerate(alpha_list):
        if (1 + idx) % 2 and alpha <= 10:
            plt.annotate(f"$\\alpha$={alpha:1.2E}", (100, max(err_map_list[idx][-1], 0.01)), fontsize=24)
        _err_mud = err_mud_list[idx]
        _err_map = err_map_list[idx]
        _err_pin = err_pin_list[idx]

        plt.plot(x, _err_mud[:-1], label='mud', c='k', lw=10)
        plt.plot(x, _err_map[:-1], label='map', c='r', ls='--', lw=5)
        plt.plot(x, _err_pin[:-1], label='lsq', c='xkcd:light blue', ls='-', lw=5)

    # plt.plot(x, regression, c='g', ls='-')
    # plt.xlim(0,dim_output)
    if 'id' in prefix:
        plt.title("Convergence for Various $\\Sigma_{init} = \\alpha I$", fontsize=1.25 * fsize)
    else:
        plt.title("Convergence for Various $\\Sigma_{init} = \\alpha \\Sigma$", fontsize=1.25 * fsize)# plt.yscale('log')
    # plt.yscale('log')
    # plt.xscale('log')
    plt.ylim(0, 1.0)
    # plt.ylim(1E-4, 5E-2)
    # plt.ylabel("$\\frac{||\\lambda^\\dagger - \\lambda||}{||\\lambda^\\dagger||}$", fontsize=fsize*1.25)
    plt.ylabel("Relative Error", fontsize=fsize * 1.25)
    plt.xlabel('Dimension of Output Space', fontsize=fsize)
    plt.legend(['mud', 'map', 'least squares'], fontsize=fsize)
    # plt.annotate(f'Slope={slope:1.4f}', (4,4/7), fontsize=32)
    plt.savefig(f'{fdir}/{prefix}-convergence.png', bbox_inches='tight')
    plt.close()
    # plt.show()

    # plt.imshow(initial_cov)

    # # ## Surface Plot

    # X, Y = np.meshgrid(x,alpha_list)
    # ZU = np.array(err_mud_list)[:,1:100]
    # ZA = np.array(err_map_list)[:,1:100]

    # # import matplotlib.pyplot as plt
    # from mpl_toolkits.mplot3d import Axes3D
    # fig = plt.figure()
    # ax = fig.add_subplot(111, projection='3d')
    # ax.plot_surface(X, np.log10(Y), ZU, alpha=0.3, color='xkcd:blue')
    # ax.plot_surface(X, np.log10(Y), ZA, alpha=0.7, color='xkcd:orange')
    # ax.set(ylabel='log10(Standard Deviation)', xlabel='Output Dimension', zlabel='Error')
    # # ax.set(yscale='log')
    # ax.view_init(15, 15)
    # plt.savefig(f'lin/{prefix}-surface-error.png', bbox_inches='tight')
    # plt.close()
    # # plt.show()


    # # # Convergence in Predictions


    # for idx, alpha in enumerate(alpha_list):
    #     _err_mud = err_Amud_list[idx]
    #     _err_map = err_Amap_list[idx]
    #     _err_pin = err_Apin_list[idx]

    #     plt.plot(np.arange(0, dim_output),_err_mud[:], label='mud', c='k', lw=10)
    #     plt.plot(np.arange(0, dim_output),_err_map[:], label='map', c='r', ls='--', lw=5)
    #     plt.plot(np.arange(0, dim_output),_err_pin[:], label='lsq', c='xkcd:light blue', ls='-', lw=5)
    # # plt.plot(x,regression, c='g', ls='-')
    # # plt.xlim(0,dim_output)
    # if 'id' in prefix:
    #     plt.title("Convergence for Various $\\Sigma_{init} = \\alpha I$", fontsize=1.25*fsize)
    # else:
    #     plt.title("Convergence for Various $\\Sigma_{init} = \\alpha \Sigma$", fontsize=1.25*fsize)# plt.yscale('log')
    # # plt.xscale('log')
    # # plt.ylim(0, 6)
    # # plt.ylim(1E-4, 5E-2)
    # plt.ylabel("Relative Error in $\mathcal{D}$", fontsize=fsize*1.25)
    # # plt.ylabel("$\\frac{||A (\\lambda^* - \\lambda) ||}{||A||}$", fontsize=fsize, fontsize=fsize*1.25)
    # plt.xlabel('Dimension of Output Space', fontsize=fsize)
    # plt.legend(['mud', 'map', 'least squares'], fontsize=fsize, loc='lower left')
    # # plt.annotate(f'Slope={slope:1.4f}', (4,4), fontsize=24)
    # plt.savefig(f'lin/{prefix}-convergence-dimension-out.png', bbox_inches='tight')
    # plt.close()
    # # plt.show()


def main_rank(args):
    """
    Main entrypoint for High-Dim Linear Rank Example
    """
    args = parse_args(args)
    setup_logging(args.loglevel)
    np.random.seed(args.seed)
#     example       = args.example
#     num_trials   = args.num_trials
#     fsize        = args.fsize
#     linewidth    = args.linewidth
#     seed         = args.seed
#     inputdim     = args.input_dim
#     save         = args.save
#     alt          = args.alt
#     bayes        = args.bayes
#     prefix       = args.prefix
#     dist         = args.dist

    presentation = False
    save = True

    if not presentation:
        plt.rcParams['mathtext.fontset'] = 'stix'
        plt.rcParams['font.family'] = 'STIXGeneral'
    fdir = 'figures/lin'
    check_dir(fdir)

    fsize = 42
    plt.figure(figsize=(10, 10))
    # ---

    # # Impact of Rank(A) for Various Choices of $\\Sigma_\text{init}$
    # We sequentially incorporate $D=1, \dots , P$ dimensions into our QoI map and study the 2-norm between the true value that was used to generate the data and the analytical MUD/MAP points. 

    dim_input, dim_output = 100, 100
    seed = 12
    np.random.seed(seed)

    # from sklearn.datasets import make_spd_matrix as make_spd
    # from sklearn.datasets import make_sparse_spd_matrix as make_cov
    # cov = np.eye(dim_input)
    initial_cov = np.diag(np.sort(np.random.rand(dim_input))[::-1] + 0.5)

    initial_mean = np.zeros(dim_input).reshape(-1, 1)
    # initial_mean = np.random.randn(dim_input).reshape(-1,1)
    randA = models.randA_list_svd
    A_list, b = models.randP(dim_input, randA=randA)
    prefix = 'lin-rank-cov'
    alpha_list = [10**(n) for n in np.linspace(-3, 4, 8)]

    # option to fix A and perturb lam_ref
    lam_ref = np.random.randn(dim_input).reshape(-1, 1)

    # d = A@lam_ref + b


    # %%time
    sols = compare_linear_sols_rank_list(lam_ref, A_list, b, alpha_list, initial_mean, initial_cov)

    # c = np.linalg.cond(A)*np.linalg.norm(lam_ref)
    c = np.linalg.norm(lam_ref)
    err_mud_list = [[np.linalg.norm(_m[0] - lam_ref) / c for _m in sols[alpha]] for alpha in alpha_list ] # output_dim+1 values of _m
    err_map_list = [[np.linalg.norm(_m[1] - lam_ref) / c for _m in sols[alpha]] for alpha in alpha_list ]
    err_pin_list = [[np.linalg.norm(_m[2] - lam_ref) / c for _m in sols[alpha]] for alpha in alpha_list ]


    err_Amud_list = [[np.linalg.norm(sum(A_list[0:i+1])@(_m[0] - lam_ref)) / np.linalg.norm(sum(A_list[0:i+1])) for i, _m in enumerate(sols[alpha])] for alpha in alpha_list ]
    err_Amap_list = [[np.linalg.norm(sum(A_list[0:i+1])@(_m[1] - lam_ref)) / np.linalg.norm(sum(A_list[0:i+1])) for i, _m in enumerate(sols[alpha])] for alpha in alpha_list ]
    err_Apin_list = [[np.linalg.norm(sum(A_list[0:i+1])@(_m[2] - lam_ref)) / np.linalg.norm(sum(A_list[0:i+1])) for i, _m in enumerate(sols[alpha])] for alpha in alpha_list ]

    # measure # of components that agree
    # err_mud_list = [[numnonzero(_m[0] - lam_ref) for _m in sols[alpha]] for alpha in alpha_list ]
    # err_map_list = [[numnonzero(_m[1] - lam_ref) for _m in sols[alpha]] for alpha in alpha_list ]
    # err_pin_list = [[numnonzero(_m[2] - lam_ref) for _m in sols[alpha]] for alpha in alpha_list ]

    # len(err_mud_list[0])
    x, y = np.arange(1, 1+dim_output, 1), err_mud_list[0]

    slope, intercept = (np.linalg.pinv(np.vander(x, 2))@np.array(y).reshape(-1,1)).ravel()
    regression = slope*x + intercept


    # # Convergence Plot

    for idx, alpha in enumerate(alpha_list):
        if (1+idx)%2 and alpha<=10:
            plt.annotate(f"$\\alpha$={alpha:1.2E}", (100, max(err_map_list[idx][-1], 0.01)), fontsize=24)
        _err_mud = err_mud_list[idx]
        _err_map = err_map_list[idx]
        _err_pin = err_pin_list[idx]

        plt.plot(x, _err_mud, label='MUD', c='k', lw=10)
        plt.plot(x, _err_map, label='MAP', c='r', ls='--', lw=5)
        plt.plot(x, _err_pin, label='LSQ', c='xkcd:light blue', ls='-', lw=5)

    # plt.plot(x, regression, c='g', ls='-')
    # plt.xlim(0,dim_output)
    if 'id' in prefix:
        plt.title("Convergence for Various $\\Sigma_{init} = \\alpha I$", fontsize=1.25*fsize)
    else:
        plt.title("Convergence for Various $\\Sigma_{init} = \\alpha \\Sigma$", fontsize=1.25*fsize)
    # plt.yscale('log')
    # plt.xscale('log')
    plt.ylim(0, 1.0)
    # plt.ylim(1E-4, 5E-2)
    plt.ylabel("$\\frac{||\\lambda^\\dagger - \\lambda||}{||\\lambda^\\dagger||}$", fontsize=fsize*1.25)
    plt.xlabel('Rank(A)', fontsize=fsize)
    plt.legend(['MUD', 'MAP', 'Least Squares'], fontsize=fsize)
    # plt.annotate(f'Slope={slope:1.4f}', (4,4/7), fontsize=32)
    plt.savefig(f'{fdir}/{prefix}-convergence.png', bbox_inches='tight')
    plt.close()
    # plt.show()

    # plt.imshow(initial_cov)


    # ---

    # ## Surface Plot

    # X, Y = np.meshgrid(x,alpha_list)
    # ZU = np.array(err_mud_list)
    # ZA = np.array(err_map_list)
    # ZI = np.array(err_pin_list)


    # # import matplotlib.pyplot as plt
    # from mpl_toolkits.mplot3d import Axes3D
    # fig = plt.figure()
    # ax = fig.add_subplot(111, projection='3d')
    # ax.plot_surface(X, np.log10(Y), ZU, alpha=0.3, color='xkcd:blue')
    # ax.plot_surface(X, np.log10(Y), ZA, alpha=0.7, color='xkcd:orange')
    # ax.set(ylabel='log10(Standard Deviation)', xlabel='Output Dimension', zlabel='Error')
    # # ax.set(yscale='log')
    # ax.view_init(15, 15)
    # # plt.savefig(f'lin/{prefix}-surface-error.png', bbox_inches='tight')
    # plt.show()

    # # print(c, slope)


    # # # Convergence in Predictions

    # for idx, alpha in enumerate(alpha_list):
    #     _err_mud = err_Amud_list[idx]
    #     _err_map = err_Amap_list[idx]
    #     _err_pin = err_Apin_list[idx]

    #     plt.plot(np.arange(0, dim_output), _err_mud[:], label='mud', c='k', lw=10)
    #     plt.plot(np.arange(0, dim_output), _err_map[:], label='map', c='r', ls='--', lw=5)
    #     plt.plot(np.arange(0, dim_output), _err_pin[:], label='lsq', c='xkcd:light blue', ls='-', lw=5)
    # # plt.plot(x,regression, c='g', ls='-')
    # # plt.xlim(0,dim_output)
    # if 'id' in prefix:
    #     plt.title("Convergence for Various $\\Sigma_{init} = \\alpha I$", fontsize=1.25*fsize)
    # else:
    #     plt.title("Convergence for Various $\\Sigma_{init} = \\alpha \Sigma$", fontsize=1.25*fsize)# plt.yscale('log')
    # plt.xscale('log')
    # plt.yscale('log')
    # # plt.ylim(0, 6)
    # # plt.ylim(1E-4, 5E-2)
    # # plt.ylabel("$\\frac{||A (\\lambda^* - \\lambda) ||}{||A||}$", fontsize=fsize)
    # plt.ylabel("Relative Error in $\mathcal{D}$", fontsize=fsize*1.25)
    # plt.xlabel('Matrix Rank', fontsize=fsize)
    # plt.legend(['mud', 'map', 'least squares'], fontsize=fsize, loc='lower left')
    # # plt.annotate(f'Slope={slope:1.4f}', (4,4), fontsize=24)
    # # plt.savefig(f'lin/{prefix}-convergence-out.png', bbox_inches='tight')
    # plt.show()


def main_contours(args):
    """
    Main entrypoint for 2D Linear Rank-Deficient Example (Contour Plots)
    """
    args = parse_args(args)
    setup_logging(args.loglevel)
    np.random.seed(args.seed)
#     example       = args.example
#     num_trials   = args.num_trials
#     fsize        = args.fsize
#     linewidth    = args.linewidth
#     seed         = args.seed
#     inputdim     = args.input_dim
#     save         = args.save
#     alt          = args.alt
#     bayes        = args.bayes
#     prefix       = args.prefix
#     dist         = args.dist

    presentation = False
    save = True

    if not presentation:
        plt.rcParams['mathtext.fontset'] = 'stix'
        plt.rcParams['font.family'] = 'STIXGeneral'
    fdir = 'figures/contours'
    check_dir(fdir)
    lam_true = np.array([0.7, 0.3])
    initial_mean = np.array([0.25, 0.25])
    A = np.array([[1, 1]])
    b = np.zeros((1,1))

    experiments = {}
    
    # data mismatch
    experiments['data_mismatch'] = {}
    experiments['data_mismatch']['out_file'] = f'{fdir}/data_mismatch_contour.png'
    experiments['data_mismatch']['data_check'] = True
    experiments['data_mismatch']['full_check'] = False
    experiments['data_mismatch']['tk_reg'] = 0
    experiments['data_mismatch']['pr_reg'] = 0
    
    # tikonov regularization
    experiments['tikonov'] = {}
    experiments['tikonov']['out_file'] = f'{fdir}/tikonov_contour.png'
    experiments['tikonov']['tk_reg'] = 1
    experiments['tikonov']['pr_reg'] = 0
    experiments['tikonov']['data_check'] = False
    experiments['tikonov']['full_check'] = False
    
    # modified regularization
    experiments['modified'] = {}
    experiments['modified']['out_file'] = f'{fdir}/consistent_contour.png'
    experiments['modified']['tk_reg'] = 1
    experiments['modified']['pr_reg'] = 1
    experiments['modified']['data_check'] = False
    experiments['modified']['full_check'] = False
    
    # map point
    experiments['classical'] = {}
    experiments['classical']['out_file'] = f'{fdir}/classical_solution.png'
    experiments['classical']['tk_reg'] = 1
    experiments['classical']['pr_reg'] = 0
    experiments['classical']['data_check'] = True
    experiments['classical']['full_check'] = True
    
    # mud point
    experiments['consistent'] = {}
    experiments['consistent']['out_file'] = f'{fdir}/consistent_solution.png'
    experiments['consistent']['tk_reg'] = 1
    experiments['consistent']['pr_reg'] = 1
    experiments['consistent']['data_check'] = True
    experiments['consistent']['full_check'] = True

    # comparison
    experiments['compare'] = {}
    experiments['compare']['out_file'] = f'{fdir}/map_compare_contour.png'
    experiments['compare']['data_check'] = True
    experiments['compare']['full_check'] = True
    experiments['compare']['tk_reg'] = 1
    experiments['compare']['pr_reg'] = 0
    experiments['compare']['comparison'] = True
    experiments['compare']['cov_01'] = -0.5

    for ex in experiments:
        _logger.info(f"Running {ex}")
        config = experiments[ex]
        out_file = config.get('out_file', 'latest_figure.png')
        tk_reg = config.get('tk_reg', 1)
        pr_reg = config.get('pr_reg', 1)
        cov_01 = config.get('cov_01', -0.25)
        cov_11 = config.get('cov_11', 0.5)
        obs_std = config.get('obs_std', 0.5)
        full_check = config.get('full_check', True)
        data_check = config.get('data_check', True)
        numr_check = config.get('numr_check', False)
        comparison = config.get('comparison', False)

        contour_example(A=A, b=b, save=save,
                        param_ref=lam_true,
                        compare=comparison,
                        cov_01=cov_01,
                        cov_11=cov_11, 
                        initial_mean=initial_mean,
                        alpha=tk_reg,
                        omega=pr_reg,
                        show_full=full_check,
                        show_data=data_check,
                        show_est=numr_check,
                        obs_std = obs_std,
                        figname=out_file)


def main(args):
    """
    Main entrypoint for example-generation
    """
    main_contours(args)
    main_rank(args)
    main_dim(args)


def run():
    """Entry point for console_scripts
    """
    main(sys.argv[1:])

############################################################




def contour_example(A=np.array([[1, 1]]), b=np.zeros([1, 1]),  # noqa: C901
                            cov_11=0.5, cov_01=-0.25,
                            initial_mean=np.array([0.25, 0.25]),
                            alpha=1, omega=1, obs_std=1,
                            show_full=True, show_data=True,
                            show_est=False, param_ref=None, compare=False,
                            fsize=42, figname='latest_figure.png', save=False,
                            ):
    """
    alpha: float in [0, 1], weight of Tikhonov regularization
    omega: float in [0, 1], weight of Modified regularization
    """
    # mesh for plotting
    N, r = 250, 1
    X, Y, XX = make_2d_unit_mesh(N, r)
    inputs = XX

    std_of_data = [obs_std]
    obs_cov = np.diag(std_of_data)
    observed_data_mean = np.array([[1]])
    initial_cov = np.array([[1, cov_01], [cov_01, cov_11]])

    assert np.all(np.linalg.eigvals(initial_cov) > 0)

    z = full_functional(A, XX, b, initial_mean, initial_cov, observed_data_mean, observed_cov=obs_cov)
    zp = norm_predicted(A, XX, initial_mean, initial_cov)
    zi = norm_input(XX, initial_mean, initial_cov)
    zd = norm_data(A, XX, b, observed_data_mean, observed_cov=obs_cov)
    # sanity check that all arguments passed correctly:
    assert np.linalg.norm(z - zi - zd + zp) < 1E-8

    # plotting contours
    z = (alpha * zi + zd - omega * zp)
    mud_a = np.argmin(z)
    map_a = np.argmin(alpha * zi + zd)

    # get mud/map points from minimal values on mesh
    mud_pt = inputs[mud_a, :]
    map_pt = inputs[map_a, :]

    msize = 500
    ls = (np.linalg.pinv(A) @ observed_data_mean.T).ravel()
    if show_data:
        plt.contour(inputs[:, 0].reshape(N, N),
                    inputs[:, 1].reshape(N, N),
                    (zd).reshape(N, N), 25,
                    cmap=cm.viridis, alpha=0.5, vmin=0, vmax=4)
        plt.axis('equal')

        s = np.linspace(-2 * r, 2 * r, 10)

        if A.shape[0] < A.shape[1]:
            # nullspace through least-squares
            null_line = null_space(A) * s + ls.reshape(-1, 1)
            plt.plot(null_line[0, :], null_line[1, :],
                     label='Solution Contour',
                     lw=2, color='xkcd:red')
            if not show_full:
                plt.annotate('Solution Contour', (0.1, 0.9),
                             fontsize=fsize, backgroundcolor="w")

    if show_full:
        plt.contour(inputs[:, 0].reshape(N, N),
                    inputs[:, 1].reshape(N, N),
                    z.reshape(N, N), 50,
                    cmap=cm.viridis, alpha=1.0)
    elif alpha + omega > 0:
        plt.contour(inputs[:, 0].reshape(N, N),
                    inputs[:, 1].reshape(N, N),
                    (alpha * zi - omega * zp).reshape(N, N), 100,
                    cmap=cm.viridis, alpha=0.25)
    plt.axis('equal')


    if alpha + omega > 0:
        plt.scatter(initial_mean[0], initial_mean[1],
                    label='Initial Mean',
                    color='k', s=msize)
        if not show_full:
            plt.annotate('Initial Mean',
                         (initial_mean[0] + 0.001 * fsize, initial_mean[1] - 0.001 * fsize),
                         fontsize=fsize, backgroundcolor="w")
        else:
            if compare:
                plt.scatter(param_ref[0], param_ref[1],
                    label='$\\lambda^\\dagger$',
                    color='k', s=msize, marker='s')

                plt.annotate('Truth',
                         (param_ref[0] + 0.00075 * fsize, param_ref[1] + 0.00075 * fsize),
                         fontsize=fsize, backgroundcolor="w")

        show_mud = omega > 0 or compare

        if show_full:
            # scatter and line from origin to least squares
            plt.scatter(ls[0], ls[1],
                        label='Least Squares',
                        color='xkcd:blue',
                        marker='d', s=msize, zorder=10)
            plt.plot([0, ls[0]], [0, ls[1]],
                     color='xkcd:blue',
                     marker='d', lw=1, zorder=10)
            plt.annotate('Least Squares',
                         (ls[0] - 0.001 * fsize, ls[1] + 0.001 * fsize),
                         fontsize=fsize, backgroundcolor="w")

            if show_est:  # numerical solutions
                if omega > 0:
                    plt.scatter(mud_pt[0], mud_pt[1],
                                label='min: Tk - Un', color='xkcd:sky blue',
                                marker='o', s=3 * msize, zorder=10)
                if (alpha > 0 and omega != 1):
                    plt.scatter(map_pt[0], map_pt[1],
                                label='min: Tk', color='xkcd:blue',
                                marker='o', s=3 * msize, zorder=10)

            if (alpha > 0 and omega != 1):  # analytical MAP point
                map_pt_eq = map_sol(A, b, observed_data_mean,
                                    initial_mean, initial_cov,
                                    data_cov=obs_cov, w=alpha)
                plt.scatter(map_pt_eq[0], map_pt_eq[1],
                            label='MAP', color='xkcd:orange',
                            marker='x', s=msize, lw=10, zorder=10)

                if compare:  # second map point has half the regularization strength
                    plt.annotate('MAP$_{\\alpha}$',
                                 (map_pt_eq[0] - 0.004 * fsize, map_pt_eq[1] - 0.002 * fsize),
                                 fontsize=fsize, backgroundcolor="w")

                else:
                    plt.annotate('MAP$_{\\alpha}$',
                                 (map_pt_eq[0] + 0.0001 * fsize, map_pt_eq[1] - 0.002 * fsize),
                                 fontsize=fsize, backgroundcolor="w")

            if show_mud:  # analytical MUD point
                mud_pt_eq = mud_sol(A, b, observed_data_mean,
                                    initial_mean, initial_cov)
                plt.scatter(mud_pt_eq[0], mud_pt_eq[1],
                            label='MUD', color='xkcd:brown',
                            marker='*', s=2 * msize, lw=5, zorder=10)
                plt.annotate('MUD',
                             (mud_pt_eq[0] + 0.001 * fsize, mud_pt_eq[1] - 0.001 * fsize),
                             fontsize=fsize, backgroundcolor="w")

        if A.shape[0] < A.shape[1]:
            # want orthogonal nullspace, function gives one that is already normalized
            v = null_space(A @ initial_cov)
            v = v[::-1]  # in 2D, we can just swap entries and put a negative sign in front of one
            v[0] = - v[0]

            if show_full and show_mud:
                # grid search to find upper/lower bounds of line being drawn.
                # importance is the direction, image is nicer with a proper origin/termination
                s = np.linspace(-1, 1, 1000)
                new_line = (v.reshape(-1, 1) * s) + initial_mean.reshape(-1, 1)
                mx = np.argmin(np.linalg.norm(new_line - initial_mean.reshape(-1, 1), axis=0))
                mn = np.argmin(np.linalg.norm(new_line - mud_pt_eq.reshape(-1, 1), axis=0))
                plt.plot(new_line[0, mn:mx], new_line[1, mn:mx], lw=1, label='projection line', c='k')
        elif show_full:
            plt.plot([initial_mean[0], ls[0]],
                     [initial_mean[1], ls[1]],
                     lw=1, label='Projection Line', c='k')

    #     print(p)

    plt.axis('square')
    plt.axis([0, r, 0, r])
#     plt.legend(fontsize=fsize)
    plt.xticks(fontsize=0.75 * fsize)
    plt.yticks(fontsize=0.75 * fsize)
    plt.tight_layout()
    if save:
        if '/' in figname:
            fdir = '/'.join(figname.split('/')[:-1])
            check_dir(fdir)
        plt.savefig(figname, dpi=300)
        plt.close()

#     plt.title('Predicted Covariance: {}'.format((A@initial_cov@A.T).ravel() ))
    # plt.show()


def compare_mud_map_pin(A, b, d, mean, cov):
    mud_pt = mud_sol(A, b, d, mean, cov)
    map_pt = map_sol(A, b, d, mean, cov)
    pin_pt = (np.linalg.pinv(A)@(d-b)).reshape(-1,1)
    return mud_pt, map_pt, pin_pt


def transform_rank_list(lam_ref, A, b, rank):
    """
    A is a list here. We sum the first `rank` elements of it
    to return a matrix with the desired rank.
    """
    _A = sum(A[0:rank])
    _b = b
    _d = _A@lam_ref + _b
    assert np.linalg.matrix_rank(_A) == rank, "Unexpected rank mismatch"
    return _A, _b, _d


def transform_dim_out(lam_ref, A, b, dim):
    if isinstance(A, list) or isinstance(A, tuple):
        raise AttributeError("A must be a matrix, not a list or tuple.")

    _A = A[:dim, :]
    _b = b[:dim, :]
    _d = _A@lam_ref + _b
    return _A, _b, _d


def compare_linear_sols_rank_list(lam_ref, A, b,
                             alpha=1, mean=None, cov=None):
    """
    Input and output dimensions fixed, varying rank 1..dim_output.
    """
    
    return compare_linear_sols(transform_rank_list, lam_ref, A, b, alpha, mean, cov)


def compare_linear_sols_dim(lam_ref, A, b,
                            alpha=1, mean=None, cov=None):
    """
    Input dimension fixed, varying output dimension.
    """
    return compare_linear_sols(transform_dim_out, lam_ref, A, b, alpha, mean, cov)


def compare_linear_sols(transform, lam_ref, A, b,
                            alpha=1, mean=None, cov=None):
    """
    Input dimension fixed, varying according to the output
    of the anonymous function `transform`'s return.
    """
    sols = {}
    if isinstance(alpha, list) or isinstance(alpha, tuple):
        alpha_list = alpha
    else:
        alpha_list = [alpha]

    if mean is None:
        mean = np.zeros(A.shape[1])
    
    if cov is None:
        cov = np.eye(A.shape[1])

    _logger.info("alpha = {}".format(alpha_list))
    if isinstance(A, list):  # svd approach returns list
        dim_output = A[0].shape[0]
    else:
        dim_output = A.shape[0]

    for alpha in alpha_list:
        sols[alpha] = []
        for _out in range(1, dim_output+1, 1):
            _A, _b, _d = transform(lam_ref, A, b, _out)
            _mud, _map, _pin = compare_mud_map_pin(_A, _b, _d, mean, alpha*cov)
            sols[alpha].append((_mud, _map, _pin))

    return sols


if __name__ == "__main__":
    run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import List

import numpy as np


def createRandomLinearMap(dim_input, dim_output,
                          dist='normal', repeated=False):
    """
    Create random linear map from P dimensions to S dimensions.
    """
    if dist == 'normal':
        M     = np.random.randn(dim_output, dim_input)  # noqa: E221
    else:
        M     = np.random.rand(dim_output, dim_input)   # noqa: E221
    if repeated:  # just use first row
        M     = np.array(list(M[0, :]) * dim_output)    # noqa: E221
        M     = M.reshape(dim_output, dim_input)        # noqa: E221

    return M


def createNoisyReferenceData(M, reference_point, std):
    dim_input  = len(reference_point)                            # noqa: E221
    dim_output = M.shape[0]
    assert M.shape[1] == dim_input, "Mperator/Data dimension mismatch"
    if isinstance(std, int) or isinstance(std, float):
        std    = np.array([std] * dim_output)                    # noqa: E221

    ref_input  = np.array(list(reference_point)).reshape(-1, 1)  # noqa: E221
    ref_data   = M @ ref_input                                   # noqa: E221
    noise      = np.diag(std) @ np.random.randn(dim_output, 1)   # noqa: E221
    data       = ref_data + noise                                # noqa: E221
    return data


def createRandomLinearPair(reference_point, num_observations, std,
                           dist='normal', repeated=False):
    """
    data will come from a normal distribution centered at zero
    with standard deviation given by `std`
    QoI map will come from standard uniform or normal if dist=normal
    if `repeated` is True, the map will be rank 1.
    """
    dim_input = len(reference_point)
    M         = createRandomLinearMap(dim_input, num_observations, dist, repeated)  # noqa: E221, E501
    data      = createNoisyReferenceData(M, reference_point, std)                   # noqa: E221, E501
    return M, data


def createRandomLinearProblem(reference_point, num_qoi,
                              num_obs_list, std_list,
                              dist='normal', repeated=False):
    """
    Wrapper around `createRandomLinearQoI` to generalize to multiple QoI maps.
    """
    if isinstance(std_list, int) or isinstance(std_list, float):
        std_list = [std_list] * num_qoi
    else:
        assert len(std_list) == num_qoi

    if isinstance(num_obs_list, int) or isinstance(num_obs_list, float):
        num_obs_list = [num_obs_list] * num_qoi
    else:
        assert len(num_obs_list) == num_qoi

    assert len(std_list) == len(num_obs_list)
    results = [createRandomLinearPair(reference_point, n, s, dist, repeated)
               for n, s in zip(num_obs_list, std_list)]
    operator_list = [r[0] for r in results]  # noqa: E221
    data_list     = [r[1] for r in results]  # noqa: E221
    return operator_list, data_list, std_list


def randA_outer(dim_output, dim_input=None, seed=None):
    """
    Generate `dimension` rank-1 matrices using Gaussian entries
    to generate a vector `x` and then take outer-product with self.
    """
    if seed is not None:
        np.random.seed(seed)
    A = []
    for i in range(dim_output):
        _x = randA_gauss(dim_output, 1)
#         _x = _x / np.linalg.norm(_x) # unit vector
        _a = np.outer(_x, _x)
        A.append(_a)
    return A


def randA_list_svd(dim_output, dim_input=None, seed=None) -> List:
    """
    Generate random square Gaussian matrix, perform SVD, and
    construct rank-1 matrices from components. Return list of them.
    Sum `R` entries of this returned list to generate a rank-R matrix.
    """
    A = []
    _A = randA_gauss(dim_output, dim_input, seed=seed)
    u, s, v = np.linalg.svd(_A)
    for i in range(dim_output):
        _a = s[i]*(u[:, i].reshape(-1, 1))@v[:, i].reshape(1, -1)
        A.append(_a)
    return A


def randA_qr(dim_output, dim_input=None, seed=None):
    """
    Generate random Gaussian matrix, perform QR,
    and returns the resulting (orthogonal) Q
    """
    A = randA_gauss(dim_output, dim_input, seed)
    A, _ = np.linalg.qr(A)
    return A


def randA_gauss(dim_output, dim_input=None, seed=None):
    """
    Generate random Gaussian matrix, perform QR,
    and returns the resulting (orthogonal) Q
    """
    if seed is not None:
        np.random.seed(seed)
    if dim_input is None:
        dim_input = dim_output
    A = np.random.randn(dim_output, dim_input)
    return A


def randP(dim_output, dim_input=None, randA=randA_gauss, seed=None):
    """
    Constructs problem set
    """
    A = randA(dim_output, dim_input, seed=seed)
    b = np.random.randn(dim_output).reshape(-1, 1)
    return A, b



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging

import matplotlib
import numpy as np
from mud.funs import map_problem, mud_problem
from mud.util import std_from_equipment
from mud_examples.experiments import (experiment_equipment,
                                      experiment_measurements)
from mud_examples.summary import extract_statistics

from mud_examples.models import generate_decay_model
from mud_examples.models import \
    generate_temporal_measurements as generate_sensors_ode
from mud_examples.summary import fit_log_linear_regression
from mud_examples.plotting import plot_decay_solution

_logger = logging.getLogger(__name__)



matplotlib.rcParams['mathtext.fontset'] = 'stix'
matplotlib.rcParams['font.family'] = 'STIXGeneral'
matplotlib.backend = 'Agg'
matplotlib.rcParams['figure.figsize'] = 10,10
matplotlib.rcParams['font.size'] = 16


def main_ode(num_trials=20,
             fsize=32,
             seed=21,
             lam_true=0.5,
             domain=[[0,1]],
             tolerances=[0.1],
             time_ratios=[0.01, 1],
             alt=False, bayes=True):
    """
    
    
    >>> from mud_examples.ode import main_ode
    >>> res = main_ode(num_trials=5, time_ratios=[0.01, 0.1, 1])
    Will run simulations for %T=[0.01, 0.1, 1]
    Running example: mud
    Measurements: [2, 20, 200]
    Plotting decay solution.
    Running example: map
    Measurements: [2, 20, 200]
    Plotting decay solution.
    """
    res = []
    print(f"Will run simulations for %T={time_ratios}")
    sd_vals      = [ std_from_equipment(tolerance=tol, probability=0.99) for tol in tolerances ]
    sigma        = sd_vals[-1] # sorted, pick largest
    t_min, t_max = 1, 3
    example_list = [ 'mud' ]
    if alt:
        example_list.append('mud-alt')

    if bayes:
        example_list.append('map')

    for example in example_list:
        print(f"Running example: {example}")
        if example == 'mud-alt':
            sensors = generate_sensors_ode(measurement_hertz=200, start_time=t_min, end_time=t_max)
        else:
            sensors = generate_sensors_ode(measurement_hertz=100, start_time=t_min, end_time=t_max)

        measurements = [ int(np.floor(len(sensors)*r)) for r in time_ratios ]
        print(f"Measurements: {measurements}")
#         times        = [ sensors[m-1] for m in measurements ]
        num_measure = max(measurements)

        model    = generate_decay_model(sensors, lam_true)
        qoi_true = model()  # no args evaluates true param
        np.random.seed(seed)
        lam = np.random.rand(int(1E3)).reshape(-1,1)
        qoi = model(lam)

        if example == 'map':
            def wrapper(num_obs, sd):
                return map_problem(domain=domain, lam=lam, qoi=qoi,
                                   sd=sd, qoi_true=qoi_true, num_obs=num_obs)
        elif example == 'mud':
            def wrapper(num_obs, sd):
                return mud_problem(domain=domain, lam=lam, qoi=qoi,
                                   sd=sd, qoi_true=qoi_true, num_obs=num_obs)
        else:
            raise ValueError("Unknown example type")

        _logger.info("Increasing Measurements Quantity Study")
        experiments, solutions = experiment_measurements(num_measurements=measurements,
                                                 sd=sigma,
                                                 num_trials=num_trials,
                                                 seed=seed,
                                                 fun=wrapper)

        means, variances = extract_statistics(solutions, lam_true)
        regression_mean, slope_mean = fit_log_linear_regression(measurements, means)
        regression_vars, slope_vars = fit_log_linear_regression(measurements, variances)

        ##########

        num_sensors = num_measure
        if len(tolerances) > 1:
            
            _logger.info("Increasing Measurement Precision Study")
            sd_means, sd_vars = experiment_equipment(num_trials=num_trials,
                                                  num_measure=num_sensors,
                                                  sd_vals=sd_vals,
                                                  reference_value=lam_true,
                                                  fun=wrapper)

            regression_err_mean, slope_err_mean = fit_log_linear_regression(tolerances, sd_means)
            regression_err_vars, slope_err_vars = fit_log_linear_regression(tolerances, sd_vars)
            _re = (regression_err_mean, slope_err_mean,
                   regression_err_vars, slope_err_vars,
                   sd_means, sd_vars, num_sensors)
        else:
            _re = None  # hack to avoid changing data structures for the time being

        # TO DO clean all this up
        _in = (lam, qoi, sensors, qoi_true, experiments, solutions)
        _rm = (regression_mean, slope_mean, regression_vars, slope_vars, means, variances)
        example_name = example.upper()
        res.append((example_name, _in, _rm, _re))

        # TODO check for existence of save directory, grab subset of measurements properly.
        plot_decay_solution(solutions, generate_decay_model, fsize=fsize,
                            end_time=t_max, lam_true=lam_true, qoi_true=qoi_true,
                            sigma=sigma, time_vector=sensors, prefix='figures/ode/' + example)

    return res


if __name__ == '__main__':
    main_ode()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging

import matplotlib.gridspec as gridspec
import numpy as np
from matplotlib import pyplot as plt
from mud_examples.utils import check_dir
from scipy.linalg import null_space

plt.rcParams['figure.figsize'] = 10, 10
plt.rcParams['font.size'] = 16
# import matplotlib
# matplotlib.rcParams['mathtext.fontset'] = 'stix'
# matplotlib.rcParams['font.family'] = 'STIXGeneral'
# matplotlib.backend = 'Agg'


_logger = logging.getLogger(__name__) # TODO: make use of this instead of print
_mpl_logger = logging.getLogger('matplotlib')
_mpl_logger.setLevel(logging.WARNING)


def plot_decay_solution(solutions, model_generator, sigma, prefix,
                        time_vector, lam_true, qoi_true, end_time=3, fsize=32, save=True):
    alpha_signal = 0.2
    alpha_points = 0.6
#     num_meas_plot_list = [25, 50, 400]
    fdir = '/'.join(prefix.split('/')[:-1])
    check_dir(fdir)
    print("Plotting decay solution.")
    for num_meas_plot in solutions:
        filename = f'{prefix}_{num_meas_plot}_reference_solution.png'
        plt.rcParams['figure.figsize'] = 25, 10
        _ = plt.figure()  # TODO: proper figure handling with `fig`

        plotting_mesh = np.linspace(0, end_time, 1000 * end_time)
        plot_model = model_generator(plotting_mesh, lam_true)
        true_response = plot_model()  # no args evaluates true param

        # true signal
        plt.plot(plotting_mesh, true_response, lw=5, c='k', alpha=1, label="True Signal, $\\xi \\sim N(0, \\sigma^2)$")

        # observations
        np.random.seed(11)
        annotate_height = 0.82
        u = qoi_true + np.random.randn(len(qoi_true)) * sigma
        plot_num_measure = num_meas_plot
        plt.scatter(time_vector[:plot_num_measure], u[:plot_num_measure], color='k', marker='.', s=250, alpha=alpha_points, label=f'{num_meas_plot} Sample Measurements')
        plt.annotate("$ \\downarrow$ Observations begin", (0.95, annotate_height), fontsize=fsize)
    #     plt.annotate("$\\downarrow$ Possible Signals", (0,annotate_height), fontsize=fsize)

        # sample signals
        num_sample_signals  = 100
        alpha_signal_sample = 0.15
        alpha_signal_mudpts = 0.45
        _true_response = plot_model(np.random.rand())  # uniform(0,1) draws from parameter space
        plt.plot(plotting_mesh, _true_response, lw=2, c='k', alpha=alpha_signal_sample, label='Predictions from Initial Density')
        for i in range(1, num_sample_signals):
            _true_response = plot_model(np.random.rand())  # uniform(0,1) draws from parameter space
            plt.plot(plotting_mesh, _true_response, lw=1, c='k', alpha=alpha_signal_sample)

        # error bars
        sigma_label = f"$\\pm3\\sigma \\qquad\\qquad \\sigma^2={sigma**2:1.3E}$"
        plt.plot(plotting_mesh[1000:], true_response[1000:] + 3 * sigma, ls='--', lw=3, c='xkcd:black', alpha=1)
        plt.plot(plotting_mesh[1000:], true_response[1000:] - 3 * sigma, ls='--', lw=3, c='xkcd:black', alpha=1, label=sigma_label)
        plt.plot(plotting_mesh[:1000], true_response[:1000] + 3 * sigma, ls='--', lw=3, c='xkcd:black', alpha=alpha_signal)
        plt.plot(plotting_mesh[:1000], true_response[:1000] - 3 * sigma, ls='--', lw=3, c='xkcd:black', alpha=alpha_signal)

        # solutions / samples
        mud_solutions = solutions[num_meas_plot]
        plt.plot(plotting_mesh, plot_model(mud_solutions[0][0]), lw=3, c='xkcd:bright red', alpha=alpha_signal_mudpts, label=f'{len(mud_solutions)} Estimates with $N={num_meas_plot:3d}$')
        for _lam in mud_solutions[1:]:
            _true_response = plot_model(_lam[0])
            plt.plot(plotting_mesh, _true_response, lw=3, c='xkcd:bright red', alpha=alpha_signal_mudpts)

        plt.ylim([0, 0.9])
        plt.xlim([0, end_time + .05])
        plt.ylabel('Response', fontsize=60)
        plt.xlabel('Time', fontsize=60)
        plt.xticks(fontsize=fsize)
        plt.yticks(fontsize=fsize)
        # legend ordering has a mind of its own, so we format it to our will
        # plt.legend(fontsize=fsize, loc='upper right')
        handles, labels = plt.gca().get_legend_handles_labels()
        order = [4, 0, 2, 1, 3]
        plt.legend([handles[idx] for idx in order], [labels[idx] for idx in order], fontsize=fsize, loc='upper right')
        plt.tight_layout()
        if save:
            plt.savefig(filename, bbox_inches='tight')
        # plt.show()



def plot_scalar_poisson_summary(res, measurements, prefix, lam_true, fsize=32, save=False):
    from fenics import plot as _plot
    from mud_examples.poisson import poissonModel # function evaluation (full response surface)

    _logger.info("Fenics plotting for 1D example: Plotting surface...")
    for _res in res:
        _prefix, _in, _rm, _re = _res
        lam, qoi, sensors, qoi_true, experiments, solutions = _in
        gamma = lam
        plot_num_measure = min(100, max(measurements))
        raveled_input = np.repeat(gamma, qoi.shape[1])
        raveled_output = qoi.reshape(-1)
        x = raveled_input
        y = raveled_output

        fig = plt.figure(figsize=(10,8))
        gs = gridspec.GridSpec(3, 3)
        ax_main = plt.subplot(gs[1:3, :2])
        # ax_xDist = plt.subplot(gs[0, :2],sharex=ax_main)
        ax_yDist = plt.subplot(gs[1:3, 2],sharey=ax_main)

        a = np.argsort(gamma)
        slopes = []

        # ax_main.plot(x,y,marker='.')
        for idx in range(plot_num_measure):
            ax_main.plot(gamma[a], qoi[a,idx], c='k',
                     label=f'sensor {idx}: (%.2f, %.2f)'%(sensors[idx,0], sensors[idx,1]),
                     lw=1, alpha=0.1)
            slopes.append(qoi[a[-1],idx] - qoi[a[0],idx])
        sa = np.argsort(slopes)
        slopes = np.array(slopes)
        ranked_slopes = slopes[sa]

        xlabel_text = "$\\lambda$"
        # ylabel_text = "$u(x_i, \lambda)$"
        ylabel_text = "Measurement\nResponse"
        ax_main.axes.set_xlabel(xlabel_text, fontsize=fsize)
        ax_main.axes.set_ylabel(ylabel_text, fontsize=fsize)
        ax_main.axes.set_ylim((-1.25,0.5))
        # ax_main.axes.set_title('Sensitivity of Measurements', fontsize=1.25*fsize)
        ax_main.axvline(3)

        ax_yDist.hist(qoi_true, bins=np.linspace(-1.25,0.5,35), orientation='horizontal', align='mid')
        # ax_yDist.set(xlabel='count')
        ax_yDist.tick_params(labelleft=False, labelbottom=False)
        if save:
            plt.savefig(f'{_prefix}_qoi_response.png', bbox_inches='tight')
        #plt.show()

        plt.figure(figsize=(10,10))
        plt.title("Sensitivity of\nMeasurement Locations", fontsize=1.25*fsize)
        plt.hist(ranked_slopes, bins=np.linspace(-1.25,0,25), density=True)
        plt.xlabel("Slope", fontsize=fsize)
        if save:
            plt.savefig(f'{_prefix}_sensitivity_qoi.png', bbox_inches='tight')
        else:
            plt.show()

        ##########

        plt.figure(figsize=(10,10))
        num_sensitive  = 20
        most_sensitive = sa[sa < 100][0:num_sensitive]
        _logger.info(f"{num_sensitive} most sensitive sensors in first 100: {most_sensitive}")
        _plot(poissonModel(lam_true))
        for i in range(min(100, max(measurements))):
            plt.scatter(sensors[i,0], sensors[i,1], c='w', s=200)
            if i in most_sensitive:
                plt.scatter(sensors[i,0], sensors[i,1], c='y', s=100)
        #     plt.annotate(f"{i+1:02d}", (sensors[i,0]-0.0125, sensors[i,1]-0.01), alpha=1, fontsize=0.35*fsize)
        # plt.title('Reference solution', fontsize=1.25*fsize)
        plt.xlabel('$x_1$', fontsize=fsize)
        plt.ylabel('$x_2$', fontsize=fsize)
        if save:
            plt.savefig(f'{_prefix}_reference_solution.png', bbox_inches='tight')
        else:
            plt.show()


def plotChain(mud_chain, ref_param, color='k', s=100):
    num_steps = len(mud_chain)
    current_point = mud_chain[0]
    plt.scatter(current_point[0], current_point[1], c='b', s=s)
    for i in range(0, num_steps):
        next_point = mud_chain[i]
        points = np.hstack([current_point, next_point])
        plt.plot(points[0, :], points[1, :], c=color)
        current_point = next_point

    plt.ylim([0, 1])
    plt.xlim([0, 1])
#     plt.axis('off')
    plt.scatter(ref_param[0], ref_param[1], c='r', s=s)


def plot_contours(A, ref_param, subset=None,
                  color='k', ls=':', lw=1, fs=20, w=1, s=100, **kwds):
    if subset is None:
        subset = np.arange(A.shape[0])
    A = A[np.array(subset), :]
    numQoI = A.shape[0]
    AA = np.hstack([null_space(A[i, :].reshape(1, -1)) for i in range(numQoI)]).T
    for i, contour in enumerate(subset):
        xloc = [ref_param[0] - w * AA[i, 0], ref_param[1] + w * AA[i, 0]]
        yloc = [ref_param[0] - w * AA[i, 1], ref_param[1] + w * AA[i, 1]]
        plt.plot(xloc, yloc, c=color, ls=ls, lw=lw, **kwds)
        plt.annotate('%d' % (contour + 1), (xloc[0], yloc[0]), fontsize=fs)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# import argparse
import logging
# import os
import sys

import matplotlib.pyplot as plt
import numpy as np
# from matplotlib import cm
from mud import __version__ as __mud_version__
from scipy.stats import \
    gaussian_kde as kde  # A standard kernel density estimator
from scipy.stats import norm, uniform  # The standard Normal distribution

from mud_examples import __version__
from mud_examples.parsers import parse_args
from mud_examples.utils import check_dir

plt.rcParams['figure.figsize'] = 10, 10
plt.rcParams['font.size'] = 24

__author__ = "Mathematical Michael"
__copyright__ = "Mathematical Michael"
__license__ = "mit"

_logger = logging.getLogger(__name__)  # TODO: use this


def setup_logging(loglevel):
    """Setup basic logging

    Args:
      loglevel (int): minimum loglevel for emitting messages
    """
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(level=loglevel, stream=sys.stdout,
                        format=logformat, datefmt="%Y-%m-%d %H:%M:%S")


def main(args):
    """
    Main entrypoint for example-generation
    """
    args = parse_args(args)
    setup_logging(args.loglevel)
    np.random.seed(args.seed)
#     example       = args.example
#     num_trials   = args.num_trials
#     fsize        = args.fsize
#     linewidth    = args.linewidth
#     seed         = args.seed
#     inputdim     = args.input_dim
#     save         = args.save
#     alt          = args.alt
#     bayes        = args.bayes
#     prefix       = args.prefix
#     dist         = args.dist
    fdir = 'figures/comparison'
    check_dir(fdir)
    presentation = False
    save = True

    if not presentation:
        plt.rcParams['mathtext.fontset'] = 'stix'
        plt.rcParams['font.family'] = 'STIXGeneral'

    # number of samples from initial and observed mean (mu) and st. dev (sigma)
    N, mu, sigma = int(1E3), 0.25, 0.1
    lam = np.random.uniform(low=-1, high=1, size=N)

    # Evaluate the QoI map on this initial sample set to form a predicted data
    qvals_predict = QoI(lam, 5)  # Evaluate lam^5 samples

    # Estimate the push-forward density for the QoI
    pi_predict = kde(qvals_predict)

    # Compute more observations for use in BIP
    tick_fsize = 28
    leg_fsize = 24
    for num_data in [1, 5, 10, 20]:
        np.random.seed(123456)  # Just for reproducibility, you can comment out if you want.
        data = norm.rvs(loc=mu, scale=sigma**2, size=num_data)

        # We will estimate the observed distribution using a parametric estimate to keep
        # the assumptions involved as similar as possible between the BIP and the SIP
        # So, we will assume the sigma is known but that the mean mu is unknown and estimated
        # from data to fit a Gaussian distribution
        mu_est = np.mean(data)

        r_approx = np.divide(norm.pdf(qvals_predict, loc=mu_est, scale=sigma), pi_predict(qvals_predict))

        # Use r to compute weighted KDE approximating the updated density
        update_kde = kde(lam, weights=r_approx)

        # Construct estimated push-forward of this updated density
        pf_update_kde = kde(qvals_predict, weights=r_approx)

        likelihood_vals = np.zeros(N)
        for i in range(N):
            likelihood_vals[i] = data_likelihood(qvals_predict[i], data,
                                                 num_data, sigma)

        # compute normalizing constants
        C_nonlinear = np.mean(likelihood_vals)
        data_like_normalized = likelihood_vals/C_nonlinear

        posterior_kde = kde(lam, weights=data_like_normalized)

        # Construct push-forward of statistical Bayesian posterior
        pf_posterior_kde = kde(qvals_predict, weights=data_like_normalized)

        fig = plt.figure()  # Plot the initial and posterior
        lam_plot = np.linspace(-1, 1, num=1000)
        plt.plot(lam_plot, uniform.pdf(lam_plot, loc=-1, scale=2), 'b--',
                 linewidth=4, label="Initial/Prior")
        plt.plot(lam_plot, update_kde(lam_plot), 'k-.',
                 linewidth=4, label="Update")
        plt.plot(lam_plot, posterior_kde(lam_plot), 'g:',
                             linewidth=4, label='Posterior')
        plt.xlim([-1, 1])
        if num_data > 1:
            plt.annotate(f'$N={num_data}$', (-0.75, 5))
            plt.ylim([0, 28])  # fix axis height for comparisons

        plt.xticks(fontsize=tick_fsize)
        plt.yticks(fontsize=tick_fsize)
        plt.xlabel("$\\Lambda$", fontsize=1.25*tick_fsize)
        plt.legend(fontsize=leg_fsize, loc='upper left')
        if save:
            plt.savefig(f'{fdir}/bip-vs-sip-{num_data}.png',
                        bbox_inches='tight')
            plt.close()
        # plt.show()

        # Plot the push-forward of the initial, observed density,
        # and push-forward of pullback and stats posterior
        plt.figure()
        qplot = np.linspace(-1, 1, num=1000)
        plt.plot(qplot, norm.pdf(qplot, loc=mu, scale=sigma), 'r-',
                 linewidth=6, label="$N(0.25,0.1^2)$")
        plt.plot(qplot, pi_predict(qplot), 'b-.',
                 linewidth=4, label="PF of Initial")
        plt.plot(qplot, pf_update_kde(qplot), 'k--',
                 linewidth=4, label="PF of Update")
        plt.plot(qplot, pf_posterior_kde(qplot), 'g:',
                 linewidth=4, label="PF of Posterior")

        plt.xlim([-1, 1])
        if num_data > 1:
            plt.annotate(f'$N={num_data}$', (-0.75, 5))
            plt.ylim([0, 20])  # fix axis height for comparisons
        plt.xticks(fontsize=tick_fsize)
        plt.yticks(fontsize=tick_fsize)
        plt.xlabel("$\\mathcal{D}$", fontsize=1.25*tick_fsize)
        plt.legend(fontsize=leg_fsize, loc='upper left')
        if save:
            plt.savefig(f'{fdir}/bip-vs-sip-pf-{num_data}.png',
                        bbox_inches='tight')
            plt.close()
        # plt.show()


def run():
    main(sys.argv[1:])


############################################################


def QoI(lam, p):
    """
    Defines a QoI mapping function as monomials to some power p
    """
    q = lam**p
    return q


def data_likelihood(qvals, data, num_data, sigma):
    v = 1.0
    for i in range(num_data):
        v *= norm.pdf(qvals-data[i], loc=0, scale=sigma)
    return v


if __name__ == '__main__':
    run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import logging
import os
import pickle
import pkgutil
import sys
from io import BytesIO
from pathlib import Path

import matplotlib.pyplot as plt  # move when migrating plotting code (maybe)
import numpy as np  # only needed for band_qoi + sample generation for main method
from mud import __version__ as __mud_version__
from mud.funs import map_problem, mud_problem
# used for convenience in setting Normal distribution
from mud.util import std_from_equipment

from mud_examples import __version__
from mud_examples.utils import LazyLoader, check_dir
from mud_examples.models import \
    generate_spatial_measurements as generate_sensors_pde

ds = LazyLoader('scipy.stats.distributions')
# from mpi4py import MPI
# comm = MPI.COMM_WORLD
# rank = comm.Get_rank()


__author__ = "Mathematical Michael"
__copyright__ = "Mathematical Michael"
__license__ = "mit"

_logger = logging.getLogger(__name__) # TODO: make use of this instead of print

try:
    fin = LazyLoader('dolfin')
except ModuleNotFoundError:
    _logger.error("Could not load fenics.")


def parse_args(args):
    """Parse command line parameters

    Args:
      args ([str]): command line parameters as list of strings

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace
    """
    parser = argparse.ArgumentParser(
        description="Poisson Problem")
    parser.add_argument(
        "--version",
        action="version",
        version=f"mud_examples {__version__}, mud {__mud_version__}")
    parser.add_argument('-n', '--num_samples',
        dest="num",
        help="Number of samples",
        default=1,
        type=int,
        metavar="INT")
    parser.add_argument('-d', '--distribution',
        dest="dist",
        help="Distribution. `n` (normal), `u` (uniform, default)",
        default='u',
        type=str,
        metavar="STR")
    parser.add_argument('-i', '--input_dim',
        dest="input_dim",
        help="Dimension of input space (default=2).",
        default=2,
        type=int,
        metavar="INT")
    parser.add_argument('-t', '--tolerance',
        dest="tolerance",
        help="Parameter for normal distribution. Proportion of samples (default: 0.95)",
        default=0.95,
        type=float,
        metavar="FLOAT")
    parser.add_argument('-b', '--beta-params',
        dest="beta_params",
        help="Parameters for beta distribution. (default = 1 1 )",
        default=None,
        nargs='+',
        type=float,
        metavar="FLOAT FLOAT")
    parser.add_argument('-p', '--prefix',
        dest="prefix",
        help="Output filename prefix (no extension)",
        default='results',
        type=str,
        metavar="STR")
    parser.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        const=logging.INFO)
    parser.add_argument(
        "-vv",
        "--very-verbose",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        const=logging.DEBUG)
    return parser.parse_args(args)


def setup_logging(loglevel):
    """Setup basic logging

    Args:
      loglevel (int): minimum loglevel for emitting messages
    """
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(level=loglevel, stream=sys.stdout,
                        format=logformat, datefmt="%Y-%m-%d %H:%M:%S")


def main(args):
    """Main entry point allowing external calls.
    Generates PDE data (requires fenics to be installed)

    Args:
      args ([str]): command line parameter list
    """
    args = parse_args(args)
    setup_logging(args.loglevel)
    _logger.debug("poisson.py data generation")

    num_samples = args.num
    dist = args.dist
    dim_input = args.input_dim
    beta_params = args.beta_params  # signals to use beta (first preference)
    tol = args.tolerance  # signals to use normal (beta must be empty), overrides uniform
    if tol < 0 or tol >= 1:
        raise ValueError("tolerance must be in (0, 1)")

    # perform random sampling according to command-line arguments
    if beta_params is None:
        if dist == 'n':  # N(-3, sd), sd chosen so 100*tol % samples are in (-4, 0)
            _logger.info(f"Generating samples from N(-3, sd), sd s.t. {100*tol}% are in (-4, 0).")
            sd = std_from_equipment(2, tol)
            randsamples = np.random.randn(num_samples, dim_input)*sd - 3
        elif dist == 'u':  # U(-4, 0)
            _logger.info(f"Generating samples from U(-4, 0).")
            randsamples = -4*np.random.rand(num_samples, dim_input)
        else:
            raise ValueError("Improper distribution choice, use `n` (normal) or `u` (uniform).")

    else:
        _logger.info("Using beta distribution since `beta_params` were passed.")
        beta_params = tuple(beta_params)
        dist = 'b' + str(beta_params).replace(', ','_').replace('(','').replace(')','')

        if len(beta_params) != 2:
            raise ValueError("Beta distribution requires only two parameters.")

        randsamples = -4*np.random.beta(*beta_params, size=(num_samples, dim_input))

    # indexed list of samples we will evaluate through our poisson model
    sample_seed_list = list(zip(range(num_samples), randsamples))

    outfile = args.prefix + str(dim_input) + str(dist)
    results = []
    for sample in sample_seed_list:
        r = evaluate_and_save_poisson(sample, outfile)
        results.append(r)
        _logger.debug(r)
    
    pickle.dump(results, open(f'{outfile}.pkl','wb'))    
    _logger.info(f"Data generation completed, saved to {outfile}.pkl.")


def run():
    """Entry point for console_scripts
    """
    main(sys.argv[1:])
    
############################################################


def poissonModel(gamma=3,
                 mesh=None, width=1,
                 nx=36, ny=36):
    """
    `gamma` is scaling parameter for left boundary condition
    `n_x` and `n_y` are the number of elements for the horizontal/vertical axes of the mesh
    """
    # Create mesh and define function space
    if mesh is None:
        mesh = fin.RectangleMesh(fin.Point(0,0), fin.Point(width, 1), nx, ny)

    V = fin.FunctionSpace(mesh, "Lagrange", 1)
    u = fin.TrialFunction(V)
    v = fin.TestFunction(V)

    # Define the left boundary condition, parameterized by gamma
    left_bc = gamma_boundary_condition(gamma)

    # Define the rest of the boundary condition
    dirichlet_bc = fin.Constant(0.0) # top and bottom
    right_bc = fin.Constant(0)

    # Define integrand
    boundary_markers = get_boundary_markers_for_rect(mesh, width)
    ds = fin.Measure('ds', domain=mesh, subdomain_data=boundary_markers)

    # Define variational problem
    f = fin.Expression("10*exp(-(pow(x[0] - 0.5, 2) + pow(x[1] - 0.5, 2)) / 0.02)", degree=2)
    a = fin.inner(fin.grad(u), fin.grad(v)) * fin.dx
    L = f * v * fin.dx + right_bc * v * ds(1) + left_bc * v * ds(3) # sum(integrals_N)

    # Define Dirichlet boundary (x = 0 or x = 1)
    def top_bottom_boundary(x):
        return x[1] < fin.DOLFIN_EPS or x[1] > 1.0 - fin.DOLFIN_EPS

    bc = fin.DirichletBC(V, dirichlet_bc, top_bottom_boundary) # this is required for correct solutions

    # Compute solution
    u = fin.Function(V)
    fin.solve(a == L, u, bc)
    return u


def gamma_boundary_condition(gamma=3):
    """
    Cannot get this to instantiate successfully in another script
    """
    if isinstance(gamma, int) or isinstance(gamma, float): # 1-D case
        # the function below will have a min at (2/7, -gamma) by design (scaling factor chosen via calculus)
        lam=gamma*823543/12500
        expr = fin.Expression(f"pow(x[1], 2) * pow(x[1] - 1, 5) * {lam}", degree=3)
    else: # Higher-D case
        expr = fin.Expression(piecewise_eval_from_vector(gamma, d=1), degree=1)
    return expr


def poisson_sensor_model(sensors, gamma, nx, ny, mesh=None):
    """
    Convenience function wrapper to just return a qoi given a parameter.
    """
    assert sensors.shape[1] == 2, "pass with shape (num_sensors, 2)"
    u = poissonModel(gamma=gamma, mesh=mesh, nx=nx, ny=ny)
    return [u(xi,yi) for xi,yi in sensors]


def eval_boundary_piecewise(u, n, d=1):
    """
    Takes an Expression `u` (on unit domain) and returns the string
    for another expression based on evaluating a piecewise-linear approximation.
    The mesh is equispaced into n intervals.
    """
    dx=1/(n+1)
    intervals = [i*dx for i in range(n+2)]
    node_values = [u(0,i) for i in intervals]
    return piecewise_eval(intervals, node_values, d)


def piecewise_eval_from_vector(u, d=1):
    """
    Takes an iterable `u` with y-values (on interior of equispaced unit domain)
    and returns the string for an expression
    based on evaluating a piecewise-linear approximation through these points.
    """
    n = len(u)
    dx = 1/(n+1)
    intervals = [i*dx for i in range(n+2)]
    node_values = [0] + list(u) + [1]
    return piecewise_eval(intervals, node_values, d)


def piecewise_eval(xvals, yvals, d=1):
    s = ''
    for i in range(1,len(xvals)):
        start = xvals[i-1]
        end = xvals[i]
        diff = start-end
        s += f' ((x[{d}] >= {start}) && (x[{d}] < {end}))*'
        s += f'({yvals[i-1]}*((x[{d}]-{end})/{diff}) + (1 - ((x[{d}]-{end})/{diff}))*{yvals[i]} ) +'
    return s[1:-1]


def eval_boundary(u, n):
    dx=1/(n+1)
    invals = [i*dx for i in range(n+2)]
    outvals = [u(0,i) for i in invals][1:-1]
    return invals[1:-1], outvals


def expressionNorm(u,v,n=100):
    u = eval_boundary(u, n)[1] 
    v = eval_boundary(v, n)[1]
    return np.linalg.norm(np.array(u) - np.array(v))/n


def copy_expression(expression):
    u = expression
    return fin.Expression(u._cppcode, **u._user_parameters,
                          degree=u.ufl_element().degree())


def get_boundary_markers_for_rect(mesh, width=1):

    class BoundaryX0(fin.SubDomain):
        def inside(self, x, on_boundary):
            return on_boundary and fin.near(x[0], 0, 1E-14)

    class BoundaryX1(fin.SubDomain):
        def inside(self, x, on_boundary):
            return on_boundary and fin.near(x[0], width, 1E-14)

    class BoundaryY0(fin.SubDomain):
        def inside(self, x, on_boundary):
            return on_boundary and fin.near(x[1], 0, 1E-14)

    class BoundaryY1(fin.SubDomain):
        def inside(self, x, on_boundary):
            return on_boundary and fin.near(x[1], 1, 1E-14)

    # not sure what the first argument here does.
    boundary_markers = fin.MeshFunction("size_t", mesh, mesh.topology().dim()-1, 0)

    # starting from top of square, going clockwise
    # we have to instantiate a class for each boundary portion in order to mark them.
    # each operation changes the state of `boundary_markers`
    BoundaryY1().mark(boundary_markers, 0)
    BoundaryX1().mark(boundary_markers, 1)
    BoundaryY0().mark(boundary_markers, 2)
    BoundaryX0().mark(boundary_markers, 3)

    return boundary_markers


def make_reproducible_without_fenics(example, lam_true=3, input_dim=2,
                                     sample_dist='u',
                                     num_samples=None, num_measure=100):
    """
    (Currently) requires XML data to be on disk, simulates sensors
    and saves everything required to one pickle file.
    """
    prefix = 'results'
    _logger.info("Running make_reproducible without fenics")
    # Either load or generate the data.
    try:  # TODO: generalize this path here... take as argument
        model_list = pickle.load(open(f'{prefix}{input_dim}{dist}.pkl', 'rb'))
        if num_samples is None or num_samples > len(model_list):
            num_samples = len(model_list)

    except FileNotFoundError as e:
        _logger.error(f"make_reproducible: {e}")
        _logger.warning("Attempting data generation with system call.")
        # below has to match where we expected our git-controlled file to be... TODO: generalize to data/
        curdir = os.getcwd().split('/')[-1]
        if curdir == 'scripts':
            fpath = f'{prefix}'
        elif curdir == 'mud-examples':
            fpath = f'scripts/{prefix}'
        else:
            fpath = f'{prefix}'
        os.system(f'generate_poisson_data -v -n 10 -i {input_dim} -p {fpath} -d {dist}')
        try:
            model_list = pickle.load(open(f'{fpath}{input_dim}{dist}.pkl', 'rb'))
            if num_samples is None or num_samples > len(model_list):
                num_samples = len(model_list)
        except TypeError:
            raise ModuleNotFoundError("Try `conda install -c conda forge fenics`")
#         fname = 'scripts/pde_2D/ref_results2u.pkl'
#         
#         _logger.warn("Need to generate data first.  Run scripts/generate_pde_data.sh")
#         _logger.warn(f"Cannot find fenics runs. Attempting load of included {fname}")
#         return fname
#         raise FileNotFoundError("Need to generate data first. Run ./generate_pde_data.sh")

    fdir = f'pde_{input_dim}D'
    check_dir(fdir)

    if input_dim == 1 and 'alt' in example:  # alternative measurement locations for more sensitivity / precision
        sensors = generate_sensors_pde(num_measure, ymax=0.95, xmax=0.25)
        fname = f'{fdir}/ref_alt_{prefix}{input_dim}{dist}.pkl'
    else:
        sensors = generate_sensors_pde(num_measure, ymax=0.95, xmax=0.95)
        fname = f'{fdir}/ref_{prefix}{input_dim}{dist}.pkl'

    lam, qoi = load_poisson_from_fenics_run(sensors, model_list[0:num_samples], nx=36, ny=36)
    qoi_ref = poisson_sensor_model(sensors, gamma=lam_true, nx=36, ny=36)

    pn = poissonModel(gamma=lam_true)
    c = pn.function_space().mesh().coordinates()
    v = [pn(c[i,0], c[i,1]) for i in range(len(c))]

    g = gamma_boundary_condition(lam_true)
    g_mesh = np.linspace(0, 1, 1000)
    g_plot = [g(0,y) for y in g_mesh]
    ref = {'sensors': sensors,
           'lam': lam, 
           'qoi': qoi, 
           'truth': lam_true, 
           'data': qoi_ref,
           'plot_u': (c,v),
           'plot_g': (g_mesh, g_plot)
          }

    with open(fname, 'wb') as f:
        pickle.dump(ref, f)
    _logger.info(fname + ' saved: ' + str(Path(fname).stat().st_size // 1000) + 'KB')

    return fname


def plot_without_fenics(fname, num_sensors=None, num_qoi=2, mode='sca', fsize = 36, example=None):
    plt.figure(figsize=(10,10))
    mode = mode.lower()
    colors = ['xkcd:red', 'xkcd:black', 'xkcd:orange', 'xkcd:blue', 'xkcd:green']

    if 'data' in fname:  # TODO turn into function.
        _logger.info(f"Loading {fname} from package")
        data = pkgutil.get_data(__package__, fname)
        data = BytesIO(data)
    else:
        _logger.info("Loading from disk")
        data = open(fname, 'rb')
    ref = pickle.load(data)

    sensors = ref['sensors']
    # qoi_ref = ref['data']
    coords, vals = ref['plot_u']
#     try:
#         import fenics as fin
#         from poisson import poissonModel
#         pn = poissonModel()
#         fin.plot(pn, vmin=-0.5, vmax=0)
#     except:
    plt.tricontourf(coords[:, 0], coords[:, 1], vals, levels=20, vmin=-0.5, vmax=0)

    # input_dim = ref['lam'].shape[1]
    
    plt.title(f"Response Surface", fontsize=1.25*fsize)
    if num_sensors is not None:  # plot sensors
        intervals = np.linspace(0, 1, num_qoi+2)[1:-1]
        if mode == 'sca':
            qoi_indices = band_qoi(sensors, 1, axis=1)
            _intervals = np.array(intervals[1:]) + \
                           ( np.array(intervals[:-1]) - \
                           np.array(intervals[1:]) ) / 2

        elif mode == 'hor':
            qoi_indices = band_qoi(sensors, num_qoi, axis=1)
            # partitions equidistant between sensors
            _intervals = np.array(intervals[1:]) + \
                           ( np.array(intervals[:-1]) - \
                           np.array(intervals[1:]) ) / 2

        elif mode == 'ver':
            qoi_indices = band_qoi(sensors, num_qoi, axis=0)
            # partitions equidistant on x_1 = (0, 1)
            _intervals = np.linspace(0, 1, num_qoi+1)[1:]
        else:
            raise ValueError("Unsupported mode type. Select from ('sca', 'ver', 'hor'). ")
        for i in range(0, len(qoi_indices)):
            _q = qoi_indices[i][qoi_indices[i] < num_sensors ]
            plt.scatter(sensors[_q,0], sensors[_q,1], s=100, color=colors[i%2])
            if i < num_qoi - 1:
                if mode == 'hor':
                    plt.axhline(_intervals[i], lw=3, c='k')
                elif mode == 'ver':
                    plt.axvline(_intervals[i], lw=3, c='k')

        plt.scatter([0]*num_qoi, intervals, s=500, marker='^', c='w')
    plt.xlim(0,1)
    plt.ylim(0,1)
#     plt.xticks([])
#     plt.yticks([])
    plt.xlabel("$x_1$", fontsize=fsize)
    plt.ylabel("$x_2$", fontsize=fsize)

    if example:
        # if 'data' in fname:  # TODO: clean this up
        #     fdir= '/'.join(fname.split('/')[1:-1])
        # else:
        #     fdir= '/'.join(fname.split('/')[:-1])
        fdir = 'figures/' + fname.replace('.pkl','/')
        # print(fdir)
        check_dir(fdir)
        fname = f"{fdir}/{example}_surface.png"
        plt.savefig(fname, bbox_inches='tight')
        _logger.info(f"Saved {fname}")


# from scipy.stats import gaussian_kde as gkde
# from scipy.stats import distributions as dist

# def ratio_dci_sing(qoi):
#     kde = gkde(qoi.T)
#     ratio_eval = dist.norm.pdf(qoi)/kde.pdf(qoi.T).ravel()
#     return ratio_eval


# def ratio_dci_mult(qois):
#     nq = np.array(qois)
#     kde = gkde(nq)
#     obs = dist.norm.pdf(nq)
#     obs_eval = np.product(obs, axis=0)
#     pre_eval = kde.pdf(nq)
#     ratio_eval = np.divide(obs_eval, pre_eval)
#     return ratio_eval


def make_mud_wrapper(domain, lam, qoi, qoi_true, indices=None, sample_dist='u', dist=ds.norm, **kwargs):
    """
    Anonymous function
    """
    if not isinstance(sample_dist, str):
        raise ValueError("`sample_dist` must be of type `str`.")
    def mud_wrapper(num_obs, sd):
        d = mud_problem(domain=domain, lam=lam, qoi=qoi, qoi_true=qoi_true, sd=sd, num_obs=num_obs, split=indices)
        d.set_initial(dist(**kwargs))
        if sample_dist == 'u':
            _logger.debug("Using weighted KDE for MUD solution.")
            d.set_predicted(weights=d._in)
        return d
    return mud_wrapper


def make_map_wrapper(domain, lam, qoi, qoi_true, log=False, dist=ds.norm, **kwargs):
    """
    Anonymous function
    """
    def map_wrapper(num_obs, sd):
        b = map_problem(domain=domain, lam=lam, qoi=qoi, qoi_true=qoi_true, sd=sd, num_obs=num_obs, log=log)
        b.set_prior(dist(**kwargs))
        return b
    return map_wrapper


# probably move to helpers or utils
def band_qoi(sensors, num_qoi=1, axis=1):
    intervals = np.linspace(0, 1, num_qoi+2)[1:-1]
    if axis == 1:
        _intervals = np.array(intervals[1:]) + ( np.array(intervals[:-1]) - np.array(intervals[1:]) ) / 2
    elif axis == 0:
        _intervals = np.linspace(0, 1, num_qoi+1)[1:]
    else:
        raise ValueError("axis must be 0 or 1 since the example is in 2D")
    _intervals = [0] + list(_intervals) + [1]
    qoi_indices = [np.where(np.logical_and(sensors[:, axis] > _intervals[i],
                                           sensors[:, axis] < _intervals[i+1]))[0] for i in range(num_qoi) ]
    return qoi_indices


def dist_from_fname(fname):
    """
    Function that infers distribution used to generate samples from the filename
    It looks for a letter before `.pkl`, i.e. `..n.pkl` -> normal distribution.
    """
    dist_from_fname = fname.strip('results').strip('res').strip('.pkl')
    dist_from_fname = dist_from_fname.split('/')[-1][-1]  # attempting to infer distribution from filename
    _logger.info(f"Inferring distribution from file name with {dist_from_fname}")
    return dist_from_fname


class pdeProblem(object):
    def __init__(self, fname=None):
        self.fname = fname
        self._lam = None
        self._lam_ref = None
        self._qoi = None
        self._qoi_ref = None
        self._sensors = None
        self._domain = None
        self._u = None
        self._g = None
        self._dist = None
        self._sample_dist = None

    @property
    def lam(self):
        return self._lam

    @lam.setter
    def lam(self, lam):
        self._lam = lam

    @property
    def lam_ref(self):
        return self._lam_ref

    @lam_ref.setter
    def lam_ref(self, lam_ref):
        if self.domain is None:
            raise AttributeError("domain not yet set.")
        min_val, max_val = 0, 4  # problem-specific
        if (lam_ref < min_val) or (lam_ref > max_val):
            raise ValueError("lam_ref must be inside domain.")
        self._lam_ref = lam_ref

    @property
    def qoi(self):
        return self._qoi

    @qoi.setter
    def qoi(self, qoi):
        self._qoi = qoi

    @property
    def qoi_ref(self):
        return self._qoi_ref

    @qoi_ref.setter
    def qoi_ref(self, qoi_ref):
        self._qoi_ref = qoi_ref

    @property
    def sensors(self):
        return self._sensors

    @sensors.setter
    def sensors(self, sensors):
        self._sensors = sensors

    @property
    def domain(self):
        return self._domain

    @domain.setter
    def domain(self, domain):
        self._domain = domain

    @property
    def g(self):
        return self._g

    @g.setter
    def g(self, g):
        self._g = g

    @property
    def u(self):
        return self._u

    @u.setter
    def u(self, u):
        self._u = u

    @property
    def dist(self):
        return self._dist

    @dist.setter
    def dist(self, dist):
        self._dist = dist

    @property
    def sample_dist(self):
        return self._sample_dist

    @sample_dist.setter
    def sample_dist(self, dist):
        if dist not in ['u', 'n']:
            raise ValueError("distribution could not be inferred. Must be from ('u', 'n')")
        self._sample_dist = dist

    def load(self, fname=None):
        if fname:
            self.fname = fname
            _logger.info(f"PDE problem loading from {fname}.")
        else:
            fname = self.fname
            _logger.info(f"PDE problem loading from default {fname}.")

        self.sample_dist = dist_from_fname(fname)

        domain, sensors, lam, qoi, qoi_ref, lam_ref, u, g = load_poisson_from_disk(fname)
        self.domain = domain
        self.sensors = sensors
        self.lam = lam
        self.lam_ref = lam_ref
        self.qoi = qoi
        self.qoi_ref = qoi_ref
        self.u = u
        self.g = g

        _logger.info(f"lam: {self.lam.shape}, qoi: {self.qoi.shape}, dist: {self.dist}")

    def map_scalar(self, log=True, **kwargs):
        _logger.info("Solving with MAP estimates.")
        return make_map_wrapper(self.domain, self.lam, self.qoi, self.qoi_ref, dist=self.dist, log=log, **kwargs)

    def mud_scalar(self, **kwargs):
        _logger.info("Solving with scalar-valued MUD estimates.")
        return make_mud_wrapper(self.domain, self.lam, self.qoi, self.qoi_ref, dist=self.dist, sample_dist=self.sample_dist, **kwargs)

    def mud_vector_horizontal(self, num_qoi=None, **kwargs):
        if num_qoi is None:  # set output dimension = input dimension
            num_qoi = self.lam.shape[1]
        indices = band_qoi(self.sensors, num_qoi=num_qoi, axis=1)
        _logger.info("Solving with horizontal-split vector-valued MUD estimates.")
        return make_mud_wrapper(self.domain, self.lam, self.qoi, self.qoi_ref, indices, dist=self.dist, sample_dist=self.sample_dist, **kwargs)

    def mud_vector_vertical(self, num_qoi=None, **kwargs):
        if num_qoi is None:  # set output dimension = input dimension
            num_qoi = self.lam.shape[1]
        indices = band_qoi(self.sensors, num_qoi=num_qoi, axis=0)
        _logger.info("Solving with vertical-split vector-valued MUD estimates.")
        return make_mud_wrapper(self.domain, self.lam, self.qoi, self.qoi_ref, indices, dist=self.dist, sample_dist=self.sample_dist, **kwargs)

    def plot_initial(self, save=True, **kwargs):
        self.plot(save=save, **kwargs)

    def plot_solutions(self, sols, num, save=True, **kwargs):
        self.plot(sols=sols, num_measurements=num, save=save, **kwargs)

    def plot(self, sols=None, num_measurements=20, example='mud', fsize=36, ftype='png', save=False):
        lam = self.lam
        qoi = self.qoi
        qoi_ref = self.qoi_ref
        # dist = self.dist
        g = self.g
        fname = self.fname.replace('.pkl', '')
        # fname = fname.replace('data/', '')
        fname = 'figures/' + fname
        check_dir(fname)
        closest_fit_index_out = np.argmin(np.linalg.norm(qoi - np.array(qoi_ref), axis=1))
        g_projected = list(lam[closest_fit_index_out, :])
        plt.figure(figsize=(10,10))

        g_mesh, g_plot = g
        intervals = list(np.linspace(0, 1, lam.shape[1]+2)[1:-1])
        # fin.plot(u_plot, mesh=mesh, lw=5, c='k', label="$g$")
        plt.plot(g_mesh, g_plot, lw=5, c='k', label="$g$")
        plt.plot([0]+intervals+[1], [0]+g_projected+[0], lw=5, c='green', alpha=0.6, ls='--', label='$\\hat{g}$', zorder=5)


        if sols is not None:
            if sols.get(num_measurements, None) is None:
                raise AttributeError(f"Solutions `sols` missing requested N={num_measurements}.")
            else:
                prefix = f'{fname}/{example}_solutions_N{num_measurements}'
                plot_lam = np.array(sols[num_measurements])
                if example == 'mud-alt':
                    qmap = '$Q_{%dD}^\\prime$'%lam.shape[1]
                    soltype = 'MUD'
                elif example == 'mud':
                    qmap = '$Q_{%dD}$'%lam.shape[1]
                    soltype = 'MUD'
                elif example == 'map':
                    qmap = '$Q_{1D}$'
                    soltype = 'MAP'
                else:
                    raise ValueError("Unsupported example type.")
                plt.title(f'{soltype} Estimates for {qmap}, $N={num_measurements}$', fontsize=1.25*fsize)
        else:  # initial plot, first 100
#             prefix = f'pde_{lam.shape[1]}{dist}/initial'
            prefix = f'{fname}/{example}_initial_S{lam.shape[0]}'
            plot_lam = lam[0:100, :]
            plt.title('Samples from Initial Density', fontsize=1.25*fsize)

        for _lam in plot_lam:
            plt.plot([0]+intervals+[1], [0]+list(_lam)+[0], lw=1, c='purple', alpha=0.2)


        plt.xlabel("$x_2$", fontsize=fsize)
        plt.ylabel("$g(x, \\lambda)$", fontsize=fsize)



        # label min(g)
        # plt.axvline(2/7, alpha=0.4, ls=':')
        # plt.axhline(-lam_true, alpha=0.4, ls=':')
        plt.ylim(-4,0)
        plt.xlim(0,1)
        plt.legend()
        if save:
            _fname = f"{prefix}.{ftype}"
            plt.savefig(_fname, bbox_inches='tight')
            _logger.info(f"Saved {_fname}")
            plt.close()
    #     plt.show()

def evaluate_and_save_poisson(sample, save_prefix):
    """
    sample is a tuple (index, gamma)
    """
    prefix = save_prefix.replace('.pkl','')
    g = sample[1]

    # define fixed mesh to avoid re-instantiation on each call to model (how it handles mesh=None)
    nx, ny = 36, 36
    mesh = fin.RectangleMesh(fin.Point(0,0), fin.Point(1,1), nx, ny)
    u = poissonModel(gamma=g, mesh=mesh, nx=nx, ny=ny)

    # Save solution as XML mesh
    fname = f"{prefix}-data/poisson-{int(sample[0]):06d}.xml"
    fin.File(fname, 'w') << u
    # tells you where to find the saved file and what the input was to generate it.
    return {int(sample[0]): {'u': fname, 'gamma': sample[1]}}

def load_poisson_from_fenics_run(sensors, file_list, nx=36, ny=36):
    num_samples = len(file_list)
    _logger.info(f"load_poisson_from_fenics_run - Loading {num_samples} evaluations of parameter space.")

    mesh = fin.RectangleMesh(fin.Point(0, 0), fin.Point(1, 1), nx, ny)
    V = fin.FunctionSpace(mesh, 'Lagrange', 1)

    qoi = []
    lam = []
    # go through all the files and load them into an array
    for i in range(num_samples):
        fname = file_list[i][i]['u']
        _logger.debug(f"Loading {fname}")
        u = fin.Function(V, fname)
        q = [u(xi, yi) for xi, yi in sensors]  # sensors
        qoi.append(np.array(q))
        lam.append(file_list[i][i]['gamma'])  # TODO: change name of this
    qoi = np.array(qoi)
    lam = np.array(lam)

    _logger.info(f'qoi: {qoi.shape}, lam: {lam.shape}, sensors: {sensors.shape}')

    return lam, qoi

def load_poisson_from_disk(fname):
    _logger.info(f"Attempting to load {fname} from disk")
    try:
        if 'data' in fname:
            _logger.info(f"Loading {fname} from package")
            data = pkgutil.get_data(__package__, fname)
            data = BytesIO(data)
        else:
            _logger.info("Loading from disk")
            data = open(fname, 'rb')
        ref = pickle.load(data)
    except FileNotFoundError:
        _logger.info(f"load_poisson_from_disk - Failed to load {fname} from disk")
        raise FileNotFoundError(f"load_poisson_from_disk: File {fname} missing. Run `make_reproducible_without_fenics` first.")
    lam = ref['lam']
    input_dim = lam.shape[1]
    domain = np.array([[-4,0]]*input_dim)
    _logger.info(f"Domain set by default as [-4, 0] for each dimension.")
    qoi = ref['qoi']
    qoi_ref = ref['data']
    lam_ref = ref['truth']
    u = ref['plot_u']
    g = ref['plot_g']
    sensors = ref['sensors']
    return domain, sensors, lam, qoi, qoi_ref, lam_ref, u, g

if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-

import unittest
import mud_examples.random as mdr
import numpy as np

__author__ = "Mathematical Michael"
__copyright__ = "Mathematical Michael"
__license__ = "mit"


class TestRandomSetups(unittest.TestCase):
    def test_random_map(self):
        for d in ['normal', 'uniform', None]:
            for r in [True, False]:
                for _dim_in in [1, 5, 10]:
                    A = mdr.createRandomLinearMap(dim_input=_dim_in,
                                                  dim_output=10,
                                                  dist=d,
                                                  repeated=r)
                    assert A is not None
                    if not r:
                        assert np.linalg.matrix_rank(A) == _dim_in
                    else:
                        assert np.linalg.matrix_rank(A) == 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import os
import types
import importlib


def check_dir(directory):
    if not os.path.exists(directory):
        os.makedirs(directory)


def make_2d_normal_mesh(N=50, window=1):
    """
    Constructs mesh based on normal distribution to
    discretize each axis.
    >>> from mud_examples.utils import make_2d_normal_mesh
    >>> x, y, XX = make_2d_normal_mesh(3)
    >>> print(XX)
    [[-1. -1.]
     [ 0. -1.]
     [ 1. -1.]
     [-1.  0.]
     [ 0.  0.]
     [ 1.  0.]
     [-1.  1.]
     [ 0.  1.]
     [ 1.  1.]]
    """
    X = np.linspace(-window, window, N)
    Y = np.linspace(-window, window, N)
    X, Y = np.meshgrid(X, Y)
    XX = np.vstack([X.ravel(), Y.ravel()]).T
    return (X, Y, XX)


def make_2d_unit_mesh(N=50, window=1):
    """
    Constructs mesh based on uniform distribution to
    discretize each axis.
    >>> from mud_examples.utils import make_2d_unit_mesh
    >>> x, y, XX = make_2d_unit_mesh(3)
    >>> print(XX)
    [[0.  0. ]
     [0.5 0. ]
     [1.  0. ]
     [0.  0.5]
     [0.5 0.5]
     [1.  0.5]
     [0.  1. ]
     [0.5 1. ]
     [1.  1. ]]
    """
    X = np.linspace(0, window, N)
    Y = np.linspace(0, window, N)
    X, Y = np.meshgrid(X, Y)
    XX = np.vstack([X.ravel(), Y.ravel()]).T
    return (X, Y, XX)


class LazyLoader(types.ModuleType):
    def __init__(self, module_name='utensor_cgen', submod_name=None):
        self._module_name = '{}{}'.format(
            module_name,
            submod_name and '.{}'.format(submod_name) or ''
        )
        self._mod = None
        super(LazyLoader, self).__init__(self._module_name)

    def _load(self):
        if self._mod is None:
            self._mod = importlib.import_module(
                self._module_name
                )
        return self._mod

    def __getattr__(self, attrb):
        try:
            return getattr(self._load(), attrb)
        except ModuleNotFoundError:
            pass

    def __dir__(self):
        return dir(self._load())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import logging
import os
import pickle
import sys
from pathlib import Path

import matplotlib

import numpy as np
from mud import __version__ as __mud_version__
from mud_examples import __version__
from mud_examples.parsers import parse_args
from mud_examples.monomial import main as main_monomial
from mud_examples.linear.lin import main as main_lin
from mud_examples.ode import main_ode
from mud_examples.pde import main_pde
from mud_examples.experiments import (plot_experiment_equipment,
                                      plot_experiment_measurements)

from mud_examples.plotting import plot_scalar_poisson_summary

matplotlib.rcParams['mathtext.fontset'] = 'stix'
matplotlib.rcParams['font.family'] = 'STIXGeneral'
matplotlib.backend = 'Agg'
matplotlib.rcParams['figure.figsize'] = 10,10
matplotlib.rcParams['font.size'] = 16


__author__ = "Mathematical Michael"
__copyright__ = "Mathematical Michael"
__license__ = "mit"

_logger = logging.getLogger(__name__)


def setup_logging(loglevel):
    """Setup basic logging

    Args:
      loglevel (int): minimum loglevel for emitting messages
    """
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(level=loglevel, stream=sys.stdout,
                        format=logformat, datefmt="%Y-%m-%d %H:%M:%S")


def main(in_args):
    """
    Main entrypoint for example-generation
    """
    args = parse_args(in_args)
    setup_logging(args.loglevel)
    np.random.seed(args.seed)
    example      = args.example
    num_trials   = args.num_trials
    fsize        = args.fsize
    linewidth    = args.linewidth
    seed         = args.seed
    inputdim     = args.input_dim
    save         = args.save
    alt          = args.alt
    bayes        = args.bayes
    sample_dist  = args.sample_dist
    dist         = args.dist
    loc          = args.loc
    scale        = args.scale

    tolerances   = list(np.sort([ float(t) for t in args.sensor_tolerance ]))

    if example == 'pde':
        measurements = list(np.sort([ int(n) for n in args.num_measure ]))
        if len(measurements) == 0:
            measurements = [100]
    else:
        time_ratios  = list(np.sort([ float(r) for r in args.ratio_measure ]))
        if len(time_ratios) == 0:
            time_ratios = [1.0]

    _logger.info("Running...")
    if example == 'pde':
        lam_true = 3.0
        res = main_pde(num_trials=num_trials,
                         fsize=fsize,
                         seed=seed,
                         lam_true=lam_true,
                         tolerances=tolerances,
                         input_dim=inputdim,
                         alt=alt, bayes=bayes,
                         dist=dist, sample_dist=sample_dist,
                         measurements=measurements,
                         loc=loc, scale=scale)

        if inputdim == 1:  # TODO: roll this plotting into main_pde, handle w/o fenics?
            plot_scalar_poisson_summary(res=res,
                                        measurements=measurements,
                                        fsize=fsize,
                                        prefix=f'figures/pde_{inputdim}D/' + example,
                                        lam_true=lam_true, save=save)
        else:
            # solution / sensors plotted by main_pde method
            pass

        if len(measurements) > 1 and inputdim == 1:
            plot_experiment_measurements(res,
                                         f'figures/pde_{inputdim}D/' + example, fsize,
                                         linewidth, save=save)

        if len(tolerances) > 1 and inputdim == 1:
            plot_experiment_equipment(tolerances, res,
                                      f'figures/pde_{inputdim}D/' + example, fsize,
                                      linewidth, save=save)
    elif example == 'ode':
        lam_true = 0.5
        res = main_ode(num_trials=num_trials,
                         fsize=fsize,
                         seed=seed,
                         lam_true=lam_true,
                         tolerances=tolerances,
                         alt=alt, bayes=bayes,
                         time_ratios=time_ratios)

        if len(time_ratios) > 1:
            plot_experiment_measurements(res,
                                         'figures/ode/' + example,
                                         fsize, linewidth,
                                         save=save, legend=True)

        if len(tolerances) > 1:
            plot_experiment_equipment(tolerances, res,
                                      'figures/ode/' + example, fsize, linewidth,
                                      title=f"Variance of MUD Error\nfor t={1+2*np.median(time_ratios):1.3f}s",
                                      save=save)

    elif example == 'lin':
        print("Running Linear Examples.")
        main_lin(in_args)

    elif example == 'monomial':
        print("Running BIP vs SIP Comparison (1D).")
        main_monomial(in_args)

    if args.save:
        with open('results.pkl', 'wb') as f:
            pickle.dump(res, f)


def run():
    """Entry point for console_scripts
    """
    main(sys.argv[1:])


def run_pde():
    """Recreates Poisson figures in MUD paper.

    >>> run_pde()
    Attempt run for measurements = [20, 100]
    Running example: mud
    Running example: map
    >>> import os; os.system('rm -rf figures/')
    0
    """
    run_cmd = """--example pde --bayes --save \
    --num-trials 20
    """.replace('    ','').replace('\n','').split(' ')
    main(run_cmd + sys.argv[1:])


def run_ode():
    """Recreates Poisson figures in MUD paper.

    >>> run_ode()
    Will run simulations for %T=[0.1, 1.0]
    Running example: mud
    Measurements: [20, 200]
    Plotting decay solution.
    Running example: map
    Measurements: [20, 200]
    Plotting decay solution.
    Plotting experiments involving increasing # of measurements.
    >>> import os; os.system('rm -rf figures/')
    0
    """
    run_cmd = """--example ode --bayes --save \
    --num-trials 20
    """.replace('    ','').replace('\n','').split(' ')
    main(run_cmd + sys.argv[1:])


def run_lin():
    """Recreates Contour figures in MUD paper.
    >>> run_lin()
    Running Linear Examples.
    >>> import os; os.system('rm -rf figures/')
    0
    """
    run_cmd = """--example lin
    """.replace('    ','').replace('\n','').split(' ')
    main(run_cmd + sys.argv[1:])


def run_monomial():
    """Recreates Contour figures in MUD paper.
    >>> run_monomial()
    Running BIP vs SIP Comparison (1D).
    >>> import os; os.system('rm -rf figures/')
    0
    """
    run_cmd = """--example monomial
    """.replace('    ','').replace('\n','').split(' ')
    main(run_cmd + sys.argv[1:])


def run_all():
    """Recreates all figures in MUD paper.
    """
    run_inv()
    run_lin()
    run_ode()
    run_pde()

############################################################


if __name__ == "__main__":
    run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import logging

from mud import __version__ as __mud_version__
from mud_examples import __version__

__author__ = "Mathematical Michael"
__copyright__ = "Mathematical Michael"
__license__ = "mit"

_logger = logging.getLogger(__name__)


def parse_args(args):
    """Parse command line parameters

    Args:
      args ([str]): command line parameters as list of strings

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace
    """

    desc = """
        Examples
        """

    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument('-e', '--example',       default='ode', type=str)
    parser.add_argument('-m', '--num-measure',   default=[20, 100],  type=int, nargs='+')
    parser.add_argument('-r', '--ratio-measure', default=[0.1, 1],  type=float, nargs='+')
    parser.add_argument('--num-trials',    default=20,    type=int)
    parser.add_argument('-t', '--sensor-tolerance',  default=[0.1], type=float, nargs='+')
    parser.add_argument('-s', '--seed',          default=21)
    parser.add_argument('-lw', '--linewidth',    default=5)
    parser.add_argument('--fsize',               default=32, type=int)
    parser.add_argument('--bayes', action='store_true')
    parser.add_argument('--alt', action='store_true')
    parser.add_argument('--save', action='store_true')

    parser.add_argument(
        "--version",
        action="version",
        version=f"mud_examples {__version__}, mud {__mud_version__}")
#     parser.add_argument('-n', '--num_samples',
#         dest="num",
#         help="Number of samples",
#         default=100,
#         type=int,
#         metavar="INT")
    parser.add_argument('-i', '--input_dim',
        dest="input_dim",
        help="Dimension of input space (default=2).",
        default=2,
        type=int,
        metavar="INT")
    parser.add_argument('-d', '--distribution',
        dest="dist",
        help="Distribution. `n` (normal), `u` (uniform, default)",
        default='u',
        type=str,
        metavar="STR")
#     parser.add_argument('-b', '--beta-params',
#         dest="beta_params",
#         help="Parameters for beta distribution. Overrides --distribution. (default = 1 1 )",
#         default=None,
#         nargs='+',
#         type=float,
#         metavar="FLOAT FLOAT")
    parser.add_argument('--loc',
        dest="loc",
        help="Prior/Initial Distribution `loc` parameter (scipy.stats.distributions).",
        default=-4.0,
        type=float,
        metavar="FLOAT")
    parser.add_argument('--scale',
        dest="scale",
        help="Prior/Initial Distribution `scale` parameter (scipy.stats.distributions).",
        default=4.0,
        type=float,
        metavar="FLOAT")
    parser.add_argument('--sample_dist',
        dest="sample_dist",
        help="Sample distribution (used for loading file) for evaluations of poisson model. default=`u`",
        default='u',
        type=str,
        metavar="STR")
    parser.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        help="set loglevel to INFO",
        action="store_const",
        const=logging.INFO)
    parser.add_argument(
        "-vv",
        "--very-verbose",
        dest="loglevel",
        help="set loglevel to DEBUG",
        action="store_const",
        const=logging.DEBUG)
    return parser.parse_args(args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import os

import matplotlib
import numpy as np
from scipy.stats import distributions as ds

import mud_examples.poisson as ps  # lazy loads fenics
from mud.funs import map_problem, mud_problem
from mud.util import std_from_equipment
from mud_examples.experiments import (experiment_equipment,
                                      experiment_measurements)
from mud_examples.summary import extract_statistics, fit_log_linear_regression
from mud_examples.utils import check_dir

_logger = logging.getLogger(__name__)


matplotlib.rcParams['mathtext.fontset'] = 'stix'
matplotlib.rcParams['font.family'] = 'STIXGeneral'
matplotlib.backend = 'Agg'
matplotlib.rcParams['figure.figsize'] = 10,10
matplotlib.rcParams['font.size'] = 16


def main_pde(num_trials=20,
             tolerances=[0.1],
             measurements=[5, 20, 50, 100, 250, 500],
             fsize=32,
             seed=21,
             lam_true=3.0,
             input_dim=2,
             dist='u', sample_dist='u',
             alt=True, bayes=True, **kwargs):
    """
    **kwargs are used for the setting of the initial distribution.
    >>> from mud_examples.pde import main_pde
    >>> res = main_pde(num_trials=10)
    Attempt run for measurements = [5, 20, 50, 100, 250, 500]
    Running example: mud
    Running example: mud-alt
    Running example: map
    """
    print(f"Attempt run for measurements = {measurements}")
    res = []
    num_measure = max(measurements)
    if sample_dist == 'n' and dist == 'u':
        raise ValueError("Weighted kde only supports uniform samples.")
    if dist == 'n':
        if 'loc' not in kwargs or 'scale' not in kwargs:
            _logger.info("Using default location/scale parameters for normal distribution")
            kwargs['loc'] = -2
            kwargs['scale'] = 0.2
        dist = ds.norm
    elif dist == 'u':
        if 'loc' not in kwargs or 'scale' not in kwargs:
            _logger.info("Using default location/scale parameters for uniform distribution")
            kwargs['loc'] = -4
            kwargs['scale'] = 4
        dist = ds.uniform
    else:  # TODO SUPPORT BETA
        raise ValueError("`dist` must be `u` or `n`")

    sd_vals     = [ std_from_equipment(tolerance=tol, probability=0.99) for tol in tolerances ]
    sigma       = sd_vals[-1] # sorted, pick largest
    _logger.info(f'Using std. dev {sigma}')
    example_list = [ 'mud' ]
    if alt:
        example_list.append('mud-alt')
    if bayes:
        example_list.append('map')

    for example in example_list:
        print(f"Running example: {example}")
        P = ps.pdeProblem()
        # in 1d this is a change in sensor location
        # in ND, change in how we partition sensors (vertical vs horizontal)
        fdir = f'pde_{input_dim}D' # expectation from make_reproducible_without_fenics

        # mud and mud alt have same sensors in higher dimensional examples
        # in 1d, the alternative approach is to change sensor placement, which requires
        # loading a separate file.
        if example == 'mud-alt' and input_dim == 1:
            fname = f'{fdir}/ref_alt_results{input_dim}{sample_dist}.pkl'
            try:
                P.load(fname)
            except FileNotFoundError:
                # attempt to load xml results from disk.
                fname = ps.make_reproducible_without_fenics('mud-alt', lam_true, input_dim=1,
                                                            num_samples=None, num_measure=num_measure,
                                                            sample_dist=sample_dist)
                P.load(fname)
            wrapper = P.mud_scalar()
            ps.plot_without_fenics(fname, num_sensors=100, num_qoi=1, example=example)
        else:
            fname = f'{fdir}/ref_results{input_dim}{sample_dist}.pkl'
            try:
                P.load(fname)
            except FileNotFoundError:
                
                try:  # available data in package
                    _logger.info("Trying packaged data.")
                    fname = 'data/' + fname
                    P.load(fname)
#                     curdir = os.getcwd().split('/')[-1]
#                     if curdir == 'scripts':
#                         raise FileNotFoundError("already within scripts directory.")
#                     _logger.warning("Attempting from scripts directory.")
#                     fname = f'scripts/{fname}'
#                     P.load(fname)
                except FileNotFoundError:
                    _logger.info("Failed to load requested data from disk or packaged datasets.")
                    fname = ps.make_reproducible_without_fenics('mud', lam_true, input_dim=input_dim,
                                                                num_samples=None, num_measure=num_measure,
                                                                sample_dist=sample_dist)
                    try:
                        P.load(fname)
                    except FileNotFoundError as e:
                        _logger.critical("Exiting program")
                        raise(e)

            P.dist = dist
            P.sample_dist = sample_dist
            # plots show only one hundred sensors to avoid clutter
            if example == 'mud-alt':
                wrapper = P.mud_vector_vertical(**kwargs)
                ps.plot_without_fenics(fname, num_sensors=100, mode='ver',
                                       num_qoi=input_dim, example=example)
            elif example == 'mud':
                wrapper = P.mud_vector_horizontal(**kwargs)
                ps.plot_without_fenics(fname, num_sensors=100, mode='hor',
                                       num_qoi=input_dim, example=example)
            elif example == 'map':
                wrapper = P.map_scalar(log=True, **kwargs)
                ps.plot_without_fenics(fname, num_sensors=100,
                                       num_qoi=input_dim, example=example)

        # adjust measurements to account for what we actually have simulated
        measurements = np.array(measurements)
        measurements = list(measurements[measurements <= P.qoi.shape[1]])
        _logger.info("Increasing Measurements Study")
        _logger.info(f"Will run simulations for N={measurements}")
        experiments, solutions = experiment_measurements(num_measurements=measurements,
                                                 sd=sigma,
                                                 num_trials=num_trials,
                                                 seed=seed,
                                                 fun=wrapper)

        means, variances = extract_statistics(solutions, lam_true)
        regression_mean, slope_mean = fit_log_linear_regression(measurements, means)
        regression_vars, slope_vars = fit_log_linear_regression(measurements, variances)

        ##########

        num_sensors = min(100, num_measure)
        if len(tolerances) > 1:
            _logger.info("Increasing Measurement Precision Study")
            sd_means, sd_vars = experiment_equipment(num_trials=num_trials,
                                                  num_measure=num_sensors,
                                                  sd_vals=sd_vals,
                                                  reference_value=lam_true,
                                                  fun=wrapper)

            regression_err_mean, slope_err_mean = fit_log_linear_regression(tolerances, sd_means)
            regression_err_vars, slope_err_vars = fit_log_linear_regression(tolerances, sd_vars)
            _re = (regression_err_mean, slope_err_mean,
                   regression_err_vars, slope_err_vars,
                   sd_means, sd_vars, num_sensors)
        else:
            _re = None  # hack to avoid changing data structures for the time being

        _in = (P.lam, P.qoi, P.sensors, P.qoi_ref, experiments, solutions)
        _rm = (regression_mean, slope_mean, regression_vars, slope_vars, means, variances)
        res.append((example, _in, _rm, _re))

        if input_dim > 1:
            if example == 'mud':
                P.plot_initial()
            for m in measurements:
                P.plot_solutions(solutions, m, example=example)
#             P.plot_solutions(solutions, 100, example=example, save=True)

    return res


if __name__ == '__main__':
    main_pde()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging

import numpy as np

_logger = logging.getLogger(__name__)


def extract_statistics(solutions, reference_value):
    """Extracts experiment statistics from solutions set
    Assumes keys of dictionary are sample sizes, and each
    value is a list containing solutions for each trial.

    >>> S = {2: [1, 1, 1], 4: [1, 1, 1]}
    >>> means, vars = extract_statistics(S, 0)
    >>> print(means)
    [1.0, 1.0]
    >>> print(vars)
    [0.0, 0.0]
    """
    num_sensors_plot_conv = solutions.keys()
    means = []
    variances = []
    for ns in num_sensors_plot_conv:
        _logger.debug(f'Extracting stats for {ns} measurements.')
        mud_solutions = solutions[ns]
        num_trials = len(mud_solutions)
        err = [np.linalg.norm(m - reference_value) for m in mud_solutions]
        assert len(err) == num_trials
        mean_mud_sol = np.mean(err)
        var_mud_sol = np.var(err)
        means.append(mean_mud_sol)
        variances.append(var_mud_sol)

    return means, variances


def fit_log_linear_regression(input_values, output_values):
    """Fits a log-linear regression

    >>> import numpy as np
    >>> x = np.arange(1,11)
    >>> np.round(fit_log_linear_regression(x,x)[1], 4)
    1.0
    """
    if len(np.unique(output_values)) == 1:
        _logger.info("Log Linear Regression: All values identical.")
        return np.array(output_values), 0
    x, y = np.log10(input_values), np.log10(output_values)
    X, Y = np.vander(x, 2), np.array(y).reshape(-1, 1)
    slope, intercept = (np.linalg.pinv(X) @ Y).ravel()
    regression_line = 10**(slope * x + intercept)
    return regression_line, slope



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np


def generate_decay_model(t, lam_true):
    def model(lam=np.array([[lam_true]])):
        if isinstance(lam, float) or isinstance(lam, int):
            lam = np.array([[lam]])
        initial_cond = 0.75
        rate = lam[:, 0].reshape(-1, 1)
        response = initial_cond * np.exp(np.outer(rate, -t))
        # this allows support for simpler 1D plotting.
        if response.shape[0] == 1:
            return response.ravel()
        return response
    return model


def generate_temporal_measurements(measurement_hertz=100,
                                   start_time=1, end_time=3):
    num_measure = measurement_hertz * (end_time - start_time)
    return np.linspace(start_time, end_time, num_measure)


def generate_spatial_measurements(num_measure,
                                  xmin=0.05, xmax=0.95,
                                  ymin=0.05, ymax=0.95):
    sensors      = np.random.rand(num_measure, 2)  # noqa: E221
    sensors[:, 0] = xmin + (xmax - xmin) * sensors[:, 0]  # x_0 location
    sensors[:, 1] = ymin + (ymax - ymin) * sensors[:, 1]  # x_1 location
    return sensors


def generate_rotation_map(qnum=10, orth=True):
    if orth:
        vec = np.linspace(0, np.pi, qnum + 1)[0:-1]
    else:
        vec = np.linspace(0, np.pi, qnum)
    matrix_entries = [[np.sin(theta), np.cos(theta)] for theta in vec]
    return np.array(matrix_entries).reshape(qnum, 2)
